__all__ = [
    "FeedForward",
    "ResidualFeedForward",
    "MultiFeedForward",
    "DeepONet",
    "MIONet",
    "FNO",
    "FourierIntegralKernel",
    "KernelNeuralOperator",
    "AveragingNeuralOperator",
    "LowRankNeuralOperator",
]

from .feed_forward import FeedForward, ResidualFeedForward
from .multi_feed_forward import MultiFeedForward
from .deeponet import DeepONet, MIONet
from .fno import FNO, FourierIntegralKernel
from .base_no import KernelNeuralOperator
from .avno import AveragingNeuralOperator
from .lno import LowRankNeuralOperator



""" Module for CausalPINN """

import torch


from torch.optim.lr_scheduler import ConstantLR

from .pinn import PINN
from pina.problem import TimeDependentProblem
from pina.utils import check_consistency


class CausalPINN(PINN):
    r"""
    Causal Physics Informed Neural Network (PINN) solver class.
    This class implements Causal Physics Informed Neural
    Network solvers, using a user specified ``model`` to solve a specific
    ``problem``. It can be used for solving both forward and inverse problems.

    The Causal Physics Informed Network aims to find
    the solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m`
    of the differential problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}

    minimizing the loss function

    .. math::
        \mathcal{L}_{\rm{problem}} = \frac{1}{N_t}\sum_{i=1}^{N_t}
        \omega_{i}\mathcal{L}_r(t_i),

    where:

    .. math::
        \mathcal{L}_r(t) = \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{A}[\mathbf{u}](\mathbf{x}_i, t)) +
        \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{B}[\mathbf{u}](\mathbf{x}_i, t))

    and,

    .. math::
        \omega_i = \exp\left(\epsilon \sum_{k=1}^{i-1}\mathcal{L}_r(t_k)\right).

    :math:`\epsilon` is an hyperparameter, default set to :math:`100`, while
    :math:`\mathcal{L}` is a specific loss function,
    default Mean Square Error:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.


    .. seealso::

        **Original reference**: Wang, Sifan, Shyam Sankaran, and Paris
        Perdikaris. "Respecting causality for training physics-informed
        neural networks." Computer Methods in Applied Mechanics
        and Engineering 421 (2024): 116813.
        DOI `10.1016 <https://doi.org/10.1016/j.cma.2024.116813>`_.

    .. note::
        This class can only work for problems inheriting
        from at least
        :class:`~pina.problem.timedep_problem.TimeDependentProblem` class.
    """

    def __init__(
        self,
        problem,
        model,
        extra_features=None,
        loss=torch.nn.MSELoss(),
        optimizer=torch.optim.Adam,
        optimizer_kwargs={"lr": 0.001},
        scheduler=ConstantLR,
        scheduler_kwargs={"factor": 1, "total_iters": 0},
        eps=100,
    ):
        """
        :param AbstractProblem problem: The formulation of the problem.
        :param torch.nn.Module model: The neural network model to use.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default :class:`torch.nn.MSELoss`.
        :param torch.nn.Module extra_features: The additional input
            features to use as augmented input.
        :param torch.optim.Optimizer optimizer: The neural network optimizer to
            use; default is :class:`torch.optim.Adam`.
        :param dict optimizer_kwargs: Optimizer constructor keyword args.
        :param torch.optim.LRScheduler scheduler: Learning
            rate scheduler.
        :param dict scheduler_kwargs: LR scheduler constructor keyword args.
        :param int | float eps: The exponential decay parameter. Note that this
            value is kept fixed during the training, but can be changed by means
            of a callback, e.g. for annealing.
        """
        super().__init__(
            problem=problem,
            model=model,
            extra_features=extra_features,
            loss=loss,
            optimizer=optimizer,
            optimizer_kwargs=optimizer_kwargs,
            scheduler=scheduler,
            scheduler_kwargs=scheduler_kwargs,
        )

        # checking consistency
        check_consistency(eps, (int, float))
        self._eps = eps
        if not isinstance(self.problem, TimeDependentProblem):
            raise ValueError(
                "Casual PINN works only for problems"
                "inheritig from TimeDependentProblem."
            )

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the Causal PINN solver based on given
        samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation
            representing the physics.
        :return: The physics loss calculated based on given
            samples and equation.
        :rtype: LabelTensor
        """
        # split sequentially ordered time tensors into chunks
        chunks, labels = self._split_tensor_into_chunks(samples)
        # compute residuals - this correspond to ordered loss functions
        # values for each time step. We apply `flatten` such that after
        # concataning the residuals we obtain a tensor of shape #chunks
        time_loss = []
        for chunk in chunks:
            chunk.labels = labels
            # classical PINN loss
            residual = self.compute_residual(samples=chunk, equation=equation)
            loss_val = self.loss(
                torch.zeros_like(residual, requires_grad=True), residual
            )
            time_loss.append(loss_val)
        # store results
        self.store_log(loss_value=float(sum(time_loss) / len(time_loss)))
        # concatenate residuals
        time_loss = torch.stack(time_loss)
        # compute weights (without the gradient storing)
        with torch.no_grad():
            weights = self._compute_weights(time_loss)
        return (weights * time_loss).mean()

    @property
    def eps(self):
        """
        The exponential decay parameter.
        """
        return self._eps

    @eps.setter
    def eps(self, value):
        """
        Setter method for the eps parameter.

        :param float value: The exponential decay parameter.
        """
        check_consistency(value, float)
        self._eps = value

    def _sort_label_tensor(self, tensor):
        """
        Sorts the label tensor based on time variables.

        :param LabelTensor tensor: The label tensor to be sorted.
        :return: The sorted label tensor based on time variables.
        :rtype: LabelTensor
        """
        # labels input tensors
        labels = tensor.labels
        # extract time tensor
        time_tensor = tensor.extract(self.problem.temporal_domain.variables)
        # sort the time tensors (this is very bad for GPU)
        _, idx = torch.sort(time_tensor.tensor.flatten())
        tensor = tensor[idx]
        tensor.labels = labels
        return tensor

    def _split_tensor_into_chunks(self, tensor):
        """
        Splits the label tensor into chunks based on time.

        :param LabelTensor tensor: The label tensor to be split.
        :return: Tuple containing the chunks and the original labels.
        :rtype: Tuple[List[LabelTensor], List]
        """
        # labels input tensors
        labels = tensor.labels
        # labels input tensors
        tensor = self._sort_label_tensor(tensor)
        # extract time tensor
        time_tensor = tensor.extract(self.problem.temporal_domain.variables)
        # count unique tensors in time
        _, idx_split = time_tensor.unique(return_counts=True)
        # splitting
        chunks = torch.split(tensor, tuple(idx_split))
        return chunks, labels  # return chunks

    def _compute_weights(self, loss):
        """
        Computes the weights for the physics loss based on the cumulative loss.

        :param LabelTensor loss: The physics loss values.
        :return: The computed weights for the physics loss.
        :rtype: LabelTensor
        """
        # compute comulative loss and multiply by epsilos
        cumulative_loss = self._eps * torch.cumsum(loss, dim=0)
        # return the exponential of the weghited negative cumulative sum
        return torch.exp(-cumulative_loss)



import torch
from copy import deepcopy

try:
    from torch.optim.lr_scheduler import LRScheduler  # torch >= 2.0
except ImportError:
    from torch.optim.lr_scheduler import (
        _LRScheduler as LRScheduler,
    )  # torch < 2.0

from .basepinn import PINNInterface
from pina.utils import check_consistency
from pina.problem import InverseProblem

from torch.optim.lr_scheduler import ConstantLR


class Weights(torch.nn.Module):
    """
    This class aims to implements the mask model for
    self adaptive weights of the Self-Adaptive
    PINN solver.
    """

    def __init__(self, func):
        """
        :param torch.nn.Module func: the mask module of SAPINN
        """
        super().__init__()
        check_consistency(func, torch.nn.Module)
        self.sa_weights = torch.nn.Parameter(torch.Tensor())
        self.func = func

    def forward(self):
        """
        Forward pass implementation for the mask module.
        It returns the function on the weights
        evaluation.

        :return: evaluation of self adaptive weights through the mask.
        :rtype: torch.Tensor
        """
        return self.func(self.sa_weights)


class SAPINN(PINNInterface):
    r"""
    Self Adaptive Physics Informed Neural Network (SAPINN) solver class.
    This class implements Self-Adaptive Physics Informed Neural
    Network solvers, using a user specified ``model`` to solve a specific
    ``problem``. It can be used for solving both forward and inverse problems.

    The Self Adapive Physics Informed Neural Network aims to find
    the solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m`
    of the differential problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}
    
    integrating the pointwise loss evaluation through a mask :math:`m` and
    self adaptive weights that permit to focus the loss function on 
    specific training samples.
    The loss function to solve the problem is

    .. math::

        \mathcal{L}_{\rm{problem}} = \frac{1}{N} \sum_{i=1}^{N_\Omega} m
        \left( \lambda_{\Omega}^{i} \right) \mathcal{L} \left( \mathcal{A}
        [\mathbf{u}](\mathbf{x}) \right) + \frac{1}{N} 
        \sum_{i=1}^{N_{\partial\Omega}}
        m \left( \lambda_{\partial\Omega}^{i} \right) \mathcal{L} 
        \left( \mathcal{B}[\mathbf{u}](\mathbf{x})
        \right),
    

    denoting the self adaptive weights as
    :math:`\lambda_{\Omega}^1, \dots, \lambda_{\Omega}^{N_\Omega}` and
    :math:`\lambda_{\partial \Omega}^1, \dots, 
    \lambda_{\Omega}^{N_\partial \Omega}`
    for :math:`\Omega` and :math:`\partial \Omega`, respectively.

    Self Adaptive Physics Informed Neural Network identifies the solution
    and appropriate self adaptive weights by solving the following problem

    .. math::

        \min_{w} \max_{\lambda_{\Omega}^k, \lambda_{\partial \Omega}^s}
        \mathcal{L} ,
    
    where :math:`w` denotes the network parameters, and
    :math:`\mathcal{L}` is a specific loss
    function, default Mean Square Error:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    .. seealso::
        **Original reference**: McClenny, Levi D., and Ulisses M. Braga-Neto.
        "Self-adaptive physics-informed neural networks."
        Journal of Computational Physics 474 (2023): 111722.
        DOI: `10.1016/
        j.jcp.2022.111722 <https://doi.org/10.1016/j.jcp.2022.111722>`_.
    """

    def __init__(
        self,
        problem,
        model,
        weights_function=torch.nn.Sigmoid(),
        extra_features=None,
        loss=torch.nn.MSELoss(),
        optimizer_model=torch.optim.Adam,
        optimizer_model_kwargs={"lr": 0.001},
        optimizer_weights=torch.optim.Adam,
        optimizer_weights_kwargs={"lr": 0.001},
        scheduler_model=ConstantLR,
        scheduler_model_kwargs={"factor": 1, "total_iters": 0},
        scheduler_weights=ConstantLR,
        scheduler_weights_kwargs={"factor": 1, "total_iters": 0},
    ):
        """
        :param AbstractProblem problem: The formualation of the problem.
        :param torch.nn.Module model: The neural network model to use
            for the model.
        :param torch.nn.Module weights_function: The neural network model
            related to the mask of SAPINN.
            default :obj:`~torch.nn.Sigmoid`.
        :param list(torch.nn.Module) extra_features: The additional input
            features to use as augmented input. If ``None`` no extra features
            are passed. If it is a list of :class:`torch.nn.Module`,
            the extra feature list is passed to all models. If it is a list
            of extra features' lists, each single list of extra feature
            is passed to a model.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default :class:`torch.nn.MSELoss`.
        :param torch.optim.Optimizer optimizer_model: The neural
            network optimizer to use for the model network
            , default is `torch.optim.Adam`.
        :param dict optimizer_model_kwargs: Optimizer constructor keyword
            args. for the model.
        :param torch.optim.Optimizer optimizer_weights: The neural
            network optimizer to use for mask model model,
            default is `torch.optim.Adam`.
        :param dict optimizer_weights_kwargs: Optimizer constructor
            keyword args. for the mask module.
        :param torch.optim.LRScheduler scheduler_model: Learning
            rate scheduler for the model.
        :param dict scheduler_model_kwargs: LR scheduler constructor
            keyword args.
        :param torch.optim.LRScheduler scheduler_weights: Learning
            rate scheduler for the mask model.
        :param dict scheduler_model_kwargs: LR scheduler constructor
            keyword args.
        """

        # check consistency weitghs_function
        check_consistency(weights_function, torch.nn.Module)

        # create models for weights
        weights_dict = {}
        for condition_name in problem.conditions:
            weights_dict[condition_name] = Weights(weights_function)
        weights_dict = torch.nn.ModuleDict(weights_dict)

        super().__init__(
            models=[model, weights_dict],
            problem=problem,
            optimizers=[optimizer_model, optimizer_weights],
            optimizers_kwargs=[
                optimizer_model_kwargs,
                optimizer_weights_kwargs,
            ],
            extra_features=extra_features,
            loss=loss,
        )

        # set automatic optimization
        self.automatic_optimization = False

        # check consistency
        check_consistency(scheduler_model, LRScheduler, subclass=True)
        check_consistency(scheduler_model_kwargs, dict)
        check_consistency(scheduler_weights, LRScheduler, subclass=True)
        check_consistency(scheduler_weights_kwargs, dict)

        # assign schedulers
        self._schedulers = [
            scheduler_model(self.optimizers[0], **scheduler_model_kwargs),
            scheduler_weights(self.optimizers[1], **scheduler_weights_kwargs),
        ]

        self._model = self.models[0]
        self._weights = self.models[1]

        self._vectorial_loss = deepcopy(loss)
        self._vectorial_loss.reduction = "none"

    def forward(self, x):
        """
        Forward pass implementation for the PINN
        solver. It returns the function
        evaluation :math:`\mathbf{u}(\mathbf{x})` at the control points
        :math:`\mathbf{x}`.

        :param LabelTensor x: Input tensor for the SAPINN solver. It expects
            a tensor :math:`N \\times D`, where :math:`N` the number of points
            in the mesh, :math:`D` the dimension of the problem,
        :return: PINN solution.
        :rtype: LabelTensor
        """
        return self.neural_net(x)

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the SAPINN solver based on given
        samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation
            representing the physics.
        :return: The physics loss calculated based on given
            samples and equation.
        :rtype: torch.Tensor
        """
        # train weights
        self.optimizer_weights.zero_grad()
        weighted_loss, _ = self._loss_phys(samples, equation)
        loss_value = -weighted_loss.as_subclass(torch.Tensor)
        self.manual_backward(loss_value)
        self.optimizer_weights.step()

        # detaching samples from the computational graph to erase it and setting
        # the gradient to true to create a new computational graph.
        # In alternative set `retain_graph=True`.
        samples = samples.detach()
        samples.requires_grad = True

        # train model
        self.optimizer_model.zero_grad()
        weighted_loss, loss = self._loss_phys(samples, equation)
        loss_value = weighted_loss.as_subclass(torch.Tensor)
        self.manual_backward(loss_value)
        self.optimizer_model.step()

        # store loss without weights
        self.store_log(loss_value=float(loss))
        return loss_value

    def loss_data(self, input_tensor, output_tensor):
        """
        Computes the data loss for the SAPINN solver based on input and
        output. It computes the loss between the
        network output against the true solution.

        :param LabelTensor input_tensor: The input to the neural networks.
        :param LabelTensor output_tensor: The true solution to compare the
            network solution.
        :return: The computed data loss.
        :rtype: torch.Tensor
        """
        # train weights
        self.optimizer_weights.zero_grad()
        weighted_loss, _ = self._loss_data(input_tensor, output_tensor)
        loss_value = -weighted_loss.as_subclass(torch.Tensor)
        self.manual_backward(loss_value)
        self.optimizer_weights.step()

        # detaching samples from the computational graph to erase it and setting
        # the gradient to true to create a new computational graph.
        # In alternative set `retain_graph=True`.
        input_tensor = input_tensor.detach()
        input_tensor.requires_grad = True

        # train model
        self.optimizer_model.zero_grad()
        weighted_loss, loss = self._loss_data(input_tensor, output_tensor)
        loss_value = weighted_loss.as_subclass(torch.Tensor)
        self.manual_backward(loss_value)
        self.optimizer_model.step()

        # store loss without weights
        self.store_log(loss_value=float(loss))
        return loss_value

    def configure_optimizers(self):
        """
        Optimizer configuration for the SAPINN
        solver.

        :return: The optimizers and the schedulers
        :rtype: tuple(list, list)
        """
        # if the problem is an InverseProblem, add the unknown parameters
        # to the parameters that the optimizer needs to optimize
        if isinstance(self.problem, InverseProblem):
            self.optimizers[0].add_param_group(
                {
                    "params": [
                        self._params[var]
                        for var in self.problem.unknown_variables
                    ]
                }
            )
        return self.optimizers, self._schedulers

    def on_train_batch_end(self, outputs, batch, batch_idx):
        """
        This method is called at the end of each training batch, and ovverides
        the PytorchLightining implementation for logging the checkpoints.

        :param torch.Tensor outputs: The output from the model for the
            current batch.
        :param tuple batch: The current batch of data.
        :param int batch_idx: The index of the current batch.
        :return: Whatever is returned by the parent
            method ``on_train_batch_end``.
        :rtype: Any
        """
        # increase by one the counter of optimization to save loggers
        self.trainer.fit_loop.epoch_loop.manual_optimization.optim_step_progress.total.completed += (
            1
        )
        return super().on_train_batch_end(outputs, batch, batch_idx)

    def on_train_start(self):
        """
        This method is called at the start of the training for setting
        the self adaptive weights as parameters of the mask model.

        :return: Whatever is returned by the parent
            method ``on_train_start``.
        :rtype: Any
        """
        device = torch.device(
            self.trainer._accelerator_connector._accelerator_flag
        )
        for condition_name, tensor in self.problem.input_pts.items():
            self.weights_dict.torchmodel[condition_name].sa_weights.data = (
                torch.rand((tensor.shape[0], 1), device=device)
            )
        return super().on_train_start()

    def on_load_checkpoint(self, checkpoint):
        """
        Overriding the Pytorch Lightning ``on_load_checkpoint`` to handle
        checkpoints for Self Adaptive Weights. This method should not be
        overridden if not intentionally.

        :param dict checkpoint: Pytorch Lightning checkpoint dict.
        """
        for condition_name, tensor in self.problem.input_pts.items():
            self.weights_dict.torchmodel[condition_name].sa_weights.data = (
                torch.rand((tensor.shape[0], 1))
            )
        return super().on_load_checkpoint(checkpoint)

    def _loss_phys(self, samples, equation):
        """
        Elaboration of the physical loss for the SAPINN solver.

        :param LabelTensor samples: Input samples to evaluate the physics loss.
        :param EquationInterface equation: the governing equation representing
            the physics.

        :return: tuple with weighted and not weighted scalar loss
        :rtype: List[LabelTensor, LabelTensor]
        """
        residual = self.compute_residual(samples, equation)
        return self._compute_loss(residual)

    def _loss_data(self, input_tensor, output_tensor):
        """
        Elaboration of the loss related to data for the SAPINN solver.

        :param LabelTensor input_tensor: The input to the neural networks.
        :param LabelTensor output_tensor: The true solution to compare the
            network solution.

        :return: tuple with weighted and not weighted scalar loss
        :rtype: List[LabelTensor, LabelTensor]
        """
        residual = self.forward(input_tensor) - output_tensor
        return self._compute_loss(residual)

    def _compute_loss(self, residual):
        """
        Elaboration of the pointwise loss through the mask model and the
        self adaptive weights

        :param LabelTensor residual: the matrix of residuals that have to
            be weighted

        :return: tuple with weighted and not weighted loss
        :rtype List[LabelTensor, LabelTensor]
        """
        weights = self.weights_dict.torchmodel[
            self.current_condition_name
        ].forward()
        loss_value = self._vectorial_loss(
            torch.zeros_like(residual, requires_grad=True), residual
        )
        return (
            self._vect_to_scalar(weights * loss_value),
            self._vect_to_scalar(loss_value),
        )

    def _vect_to_scalar(self, loss_value):
        """
        Elaboration of the pointwise loss through the mask model and the
        self adaptive weights

        :param LabelTensor loss_value: the matrix of pointwise loss

        :return: the scalar loss
        :rtype LabelTensor
        """
        if self.loss.reduction == "mean":
            ret = torch.mean(loss_value)
        elif self.loss.reduction == "sum":
            ret = torch.sum(loss_value)
        else:
            raise RuntimeError(
                f"Invalid reduction, got {self.loss.reduction} "
                "but expected mean or sum."
            )
        return ret

    @property
    def neural_net(self):
        """
        Returns the neural network model.

        :return: The neural network model.
        :rtype: torch.nn.Module
        """
        return self.models[0]

    @property
    def weights_dict(self):
        """
        Return the mask models associate to the application of
        the mask to the self adaptive weights for each loss that
        compones the global loss of the problem.

        :return: The ModuleDict for mask models.
        :rtype: torch.nn.ModuleDict
        """
        return self.models[1]

    @property
    def scheduler_model(self):
        """
        Returns the scheduler associated with the neural network model.

        :return: The scheduler for the neural network model.
        :rtype: torch.optim.lr_scheduler._LRScheduler
        """
        return self._scheduler[0]

    @property
    def scheduler_weights(self):
        """
        Returns the scheduler associated with the mask model (if applicable).

        :return: The scheduler for the mask model.
        :rtype: torch.optim.lr_scheduler._LRScheduler
        """
        return self._scheduler[1]

    @property
    def optimizer_model(self):
        """
        Returns the optimizer associated with the neural network model.

        :return: The optimizer for the neural network model.
        :rtype: torch.optim.Optimizer
        """
        return self.optimizers[0]

    @property
    def optimizer_weights(self):
        """
        Returns the optimizer associated with the mask model (if applicable).

        :return: The optimizer for the mask model.
        :rtype: torch.optim.Optimizer
        """
        return self.optimizers[1]



"""
Fourier Neural Operator Module.
"""

import torch
import torch.nn as nn
from pina import LabelTensor
import warnings
from ..utils import check_consistency
from .layers.fourier import FourierBlock1D, FourierBlock2D, FourierBlock3D
from .base_no import KernelNeuralOperator


class FourierIntegralKernel(torch.nn.Module):
    """
    Implementation of Fourier Integral Kernel network.

    This class implements the Fourier Integral Kernel network, which is a
    PINA implementation of Fourier Neural Operator kernel network.
    It performs global convolution by operating in the Fourier space.

    .. seealso::

        **Original reference**: Li, Z., Kovachki, N., Azizzadenesheli,
        K., Liu, B., Bhattacharya, K., Stuart, A., & Anandkumar, A.
        (2020). *Fourier neural operator for parametric partial
        differential equations*.
        DOI: `arXiv preprint arXiv:2010.08895.
        <https://arxiv.org/abs/2010.08895>`_
    """

    def __init__(
        self,
        input_numb_fields,
        output_numb_fields,
        n_modes,
        dimensions=3,
        padding=8,
        padding_type="constant",
        inner_size=20,
        n_layers=2,
        func=nn.Tanh,
        layers=None,
    ):
        """
        :param int input_numb_fields: Number of input fields.
        :param int output_numb_fields: Number of output fields.
        :param int | list[int] n_modes: Number of modes.
        :param int dimensions: Number of dimensions (1, 2, or 3).
        :param int padding: Padding size, defaults to 8.
        :param str padding_type: Type of padding, defaults to "constant".
        :param int inner_size: Inner size, defaults to 20.
        :param int n_layers: Number of layers, defaults to 2.
        :param torch.nn.Module func: Activation function, defaults to nn.Tanh.
        :param list[int] layers: List of layer sizes, defaults to None.
        """
        super().__init__()

        # check type consistency
        check_consistency(dimensions, int)
        check_consistency(padding, int)
        check_consistency(padding_type, str)
        check_consistency(inner_size, int)
        check_consistency(n_layers, int)
        check_consistency(func, nn.Module, subclass=True)

        if layers is not None:
            if isinstance(layers, (tuple, list)):
                check_consistency(layers, int)
            else:
                raise ValueError("layers must be tuple or list of int.")
        if not isinstance(n_modes, (list, tuple, int)):
            raise ValueError(
                "n_modes must be a int or list or tuple of valid modes."
                " More information on the official documentation."
            )

        # assign padding
        self._padding = padding

        # initialize fourier layer for each dimension
        if dimensions == 1:
            fourier_layer = FourierBlock1D
        elif dimensions == 2:
            fourier_layer = FourierBlock2D
        elif dimensions == 3:
            fourier_layer = FourierBlock3D
        else:
            raise NotImplementedError("FNO implemented only for 1D/2D/3D data.")

        # Here we build the FNO kernels by stacking Fourier Blocks

        # 1. Assign output dimensions for each FNO layer
        if layers is None:
            layers = [inner_size] * n_layers

        # 2. Assign activation functions for each FNO layer
        if isinstance(func, list):
            if len(layers) != len(func):
                raise RuntimeError(
                    "Uncosistent number of layers and functions."
                )
            _functions = func
        else:
            _functions = [func for _ in range(len(layers) - 1)]
        _functions.append(torch.nn.Identity)

        # 3. Assign modes functions for each FNO layer
        if isinstance(n_modes, list):
            if all(isinstance(i, list) for i in n_modes) and len(layers) != len(
                n_modes
            ):
                raise RuntimeError(
                    "Uncosistent number of layers and functions."
                )
            elif all(isinstance(i, int) for i in n_modes):
                n_modes = [n_modes] * len(layers)
        else:
            n_modes = [n_modes] * len(layers)

        # 4. Build the FNO network
        _layers = []
        tmp_layers = [input_numb_fields] + layers + [output_numb_fields]
        for i in range(len(layers)):
            _layers.append(
                fourier_layer(
                    input_numb_fields=tmp_layers[i],
                    output_numb_fields=tmp_layers[i + 1],
                    n_modes=n_modes[i],
                    activation=_functions[i],
                )
            )
        self._layers = nn.Sequential(*_layers)

        # 5. Padding values for spectral conv
        if isinstance(padding, int):
            padding = [padding] * dimensions
        self._ipad = [-pad if pad > 0 else None for pad in padding[:dimensions]]
        self._padding_type = padding_type
        self._pad = [
            val for pair in zip([0] * dimensions, padding) for val in pair
        ]

    def forward(self, x):
        """
        Forward computation for Fourier Neural Operator. It performs a
        lifting of the input by the ``lifting_net``. Then different layers
        of Fourier Blocks are applied. Finally the output is projected
        to the final dimensionality by the ``projecting_net``.

        :param torch.Tensor x: The input tensor for fourier block,
            depending on ``dimension`` in the initialization.
            In particular it is expected:

            * 1D tensors: ``[batch, X, channels]``
            * 2D tensors: ``[batch, X, Y, channels]``
            * 3D tensors: ``[batch, X, Y, Z, channels]``
        :return: The output tensor obtained from the kernels convolution.
        :rtype: torch.Tensor
        """
        if isinstance(x, LabelTensor):  # TODO remove when Network is fixed
            warnings.warn(
                "LabelTensor passed as input is not allowed,"
                " casting LabelTensor to Torch.Tensor"
            )
            x = x.as_subclass(torch.Tensor)
        # permuting the input [batch, channels, x, y, ...]
        permutation_idx = [0, x.ndim - 1, *[i for i in range(1, x.ndim - 1)]]
        x = x.permute(permutation_idx)

        # padding the input
        x = torch.nn.functional.pad(x, pad=self._pad, mode=self._padding_type)

        # apply fourier layers
        x = self._layers(x)

        # remove padding
        idxs = [slice(None), slice(None)] + [slice(pad) for pad in self._ipad]
        x = x[idxs]

        # permuting back [batch, x, y, ..., channels]
        permutation_idx = [0, *[i for i in range(2, x.ndim)], 1]
        x = x.permute(permutation_idx)

        return x


class FNO(KernelNeuralOperator):
    """
    The PINA implementation of Fourier Neural Operator network.

    Fourier Neural Operator (FNO) is a general architecture for
    learning Operators. Unlike traditional machine learning methods
    FNO is designed to map entire functions to other functions. It
    can be trained with Supervised learning strategies. FNO does global
    convolution by performing the operation on the Fourier space.

    .. seealso::

        **Original reference**: Li, Z., Kovachki, N., Azizzadenesheli,
        K., Liu, B., Bhattacharya, K., Stuart, A., & Anandkumar, A.
        (2020). *Fourier neural operator for parametric partial
        differential equations*.
        DOI: `arXiv preprint arXiv:2010.08895.
        <https://arxiv.org/abs/2010.08895>`_
    """

    def __init__(
        self,
        lifting_net,
        projecting_net,
        n_modes,
        dimensions=3,
        padding=8,
        padding_type="constant",
        inner_size=20,
        n_layers=2,
        func=nn.Tanh,
        layers=None,
    ):
        """
        :param torch.nn.Module lifting_net: The neural network for lifting
            the input.
        :param torch.nn.Module projecting_net: The neural network for
            projecting the output.
        :param int | list[int] n_modes: Number of modes.
        :param int dimensions: Number of dimensions (1, 2, or 3).
        :param int padding: Padding size, defaults to 8.
        :param str padding_type: Type of padding, defaults to `constant`.
        :param int inner_size: Inner size, defaults to 20.
        :param int n_layers: Number of layers, defaults to 2.
        :param torch.nn.Module func: Activation function, defaults to nn.Tanh.
        :param list[int] layers: List of layer sizes, defaults to None.
        """
        lifting_operator_out = lifting_net(
            torch.rand(size=next(lifting_net.parameters()).size())
        ).shape[-1]
        super().__init__(
            lifting_operator=lifting_net,
            projection_operator=projecting_net,
            integral_kernels=FourierIntegralKernel(
                input_numb_fields=lifting_operator_out,
                output_numb_fields=next(projecting_net.parameters()).size(),
                n_modes=n_modes,
                dimensions=dimensions,
                padding=padding,
                padding_type=padding_type,
                inner_size=inner_size,
                n_layers=n_layers,
                func=func,
                layers=layers,
            ),
        )

    def forward(self, x):
        """
        Forward computation for Fourier Neural Operator. It performs a
        lifting of the input by the ``lifting_net``. Then different layers
        of Fourier Blocks are applied. Finally the output is projected
        to the final dimensionality by the ``projecting_net``.

        :param torch.Tensor x: The input tensor for fourier block,
            depending on ``dimension`` in the initialization. In
            particular it is expected:

            * 1D tensors: ``[batch, X, channels]``
            * 2D tensors: ``[batch, X, Y, channels]``
            * 3D tensors: ``[batch, X, Y, Z, channels]``
        :return: The output tensor obtained from FNO.
        :rtype: torch.Tensor
        """
        return super().forward(x)



""" Module for Physics Informed Neural Network. """

import torch

try:
    from torch.optim.lr_scheduler import LRScheduler  # torch >= 2.0
except ImportError:
    from torch.optim.lr_scheduler import (
        _LRScheduler as LRScheduler,
    )  # torch < 2.0

from torch.optim.lr_scheduler import ConstantLR

from .basepinn import PINNInterface
from pina.utils import check_consistency
from pina.problem import InverseProblem


class PINN(PINNInterface):
    r"""
    Physics Informed Neural Network (PINN) solver class.
    This class implements Physics Informed Neural
    Network solvers, using a user specified ``model`` to solve a specific
    ``problem``. It can be used for solving both forward and inverse problems.

    The Physics Informed Network aims to find
    the solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m`
    of the differential problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}

    minimizing the loss function

    .. math::
        \mathcal{L}_{\rm{problem}} = \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{A}[\mathbf{u}](\mathbf{x}_i)) +
        \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{B}[\mathbf{u}](\mathbf{x}_i))

    where :math:`\mathcal{L}` is a specific loss function, default Mean Square Error:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    .. seealso::

        **Original reference**: Karniadakis, G. E., Kevrekidis, I. G., Lu, L.,
        Perdikaris, P., Wang, S., & Yang, L. (2021).
        Physics-informed machine learning. Nature Reviews Physics, 3, 422-440.
        DOI: `10.1038 <https://doi.org/10.1038/s42254-021-00314-5>`_.
    """

    def __init__(
        self,
        problem,
        model,
        extra_features=None,
        loss=torch.nn.MSELoss(),
        optimizer=torch.optim.Adam,
        optimizer_kwargs={"lr": 0.001},
        scheduler=ConstantLR,
        scheduler_kwargs={"factor": 1, "total_iters": 0},
    ):
        """
        :param AbstractProblem problem: The formulation of the problem.
        :param torch.nn.Module model: The neural network model to use.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default :class:`torch.nn.MSELoss`.
        :param torch.nn.Module extra_features: The additional input
            features to use as augmented input.
        :param torch.optim.Optimizer optimizer: The neural network optimizer to
            use; default is :class:`torch.optim.Adam`.
        :param dict optimizer_kwargs: Optimizer constructor keyword args.
        :param torch.optim.LRScheduler scheduler: Learning
            rate scheduler.
        :param dict scheduler_kwargs: LR scheduler constructor keyword args.
        """
        super().__init__(
            models=[model],
            problem=problem,
            optimizers=[optimizer],
            optimizers_kwargs=[optimizer_kwargs],
            extra_features=extra_features,
            loss=loss,
        )

        # check consistency
        check_consistency(scheduler, LRScheduler, subclass=True)
        check_consistency(scheduler_kwargs, dict)

        # assign variables
        self._scheduler = scheduler(self.optimizers[0], **scheduler_kwargs)
        self._neural_net = self.models[0]

    def forward(self, x):
        r"""
        Forward pass implementation for the PINN solver. It returns the function
        evaluation :math:`\mathbf{u}(\mathbf{x})` at the control points
        :math:`\mathbf{x}`.

        :param LabelTensor x: Input tensor for the PINN solver. It expects
            a tensor :math:`N \times D`, where :math:`N` the number of points
            in the mesh, :math:`D` the dimension of the problem,
        :return: PINN solution evaluated at contro points.
        :rtype: LabelTensor
        """
        return self.neural_net(x)

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the PINN solver based on given
        samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation
            representing the physics.
        :return: The physics loss calculated based on given
            samples and equation.
        :rtype: LabelTensor
        """
        residual = self.compute_residual(samples=samples, equation=equation)
        loss_value = self.loss(
            torch.zeros_like(residual, requires_grad=True), residual
        )
        self.store_log(loss_value=float(loss_value))
        return loss_value

    def configure_optimizers(self):
        """
        Optimizer configuration for the PINN
        solver.

        :return: The optimizers and the schedulers
        :rtype: tuple(list, list)
        """
        # if the problem is an InverseProblem, add the unknown parameters
        # to the parameters that the optimizer needs to optimize
        if isinstance(self.problem, InverseProblem):
            self.optimizers[0].add_param_group(
                {
                    "params": [
                        self._params[var]
                        for var in self.problem.unknown_variables
                    ]
                }
            )
        return self.optimizers, [self.scheduler]

    @property
    def scheduler(self):
        """
        Scheduler for the PINN training.
        """
        return self._scheduler

    @property
    def neural_net(self):
        """
        Neural network for the PINN training.
        """
        return self._neural_net



""" Module for SupervisedSolver """

import torch

try:
    from torch.optim.lr_scheduler import LRScheduler  # torch >= 2.0
except ImportError:
    from torch.optim.lr_scheduler import (
        _LRScheduler as LRScheduler,
    )  # torch < 2.0

from torch.optim.lr_scheduler import ConstantLR

from .solver import SolverInterface
from ..label_tensor import LabelTensor
from ..utils import check_consistency
from ..loss import LossInterface
from torch.nn.modules.loss import _Loss


class SupervisedSolver(SolverInterface):
    r"""
    SupervisedSolver solver class. This class implements a SupervisedSolver,
    using a user specified ``model`` to solve a specific ``problem``.

    The  Supervised Solver class aims to find
    a map between the input :math:`\mathbf{s}:\Omega\rightarrow\mathbb{R}^m`
    and the output :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m`. The input
    can be discretised in space (as in :obj:`~pina.solvers.rom.ROMe2eSolver`),
    or not (e.g. when training Neural Operators).

    Given a model :math:`\mathcal{M}`, the following loss function is
    minimized during training:

    .. math::
        \mathcal{L}_{\rm{problem}} = \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathbf{u}_i - \mathcal{M}(\mathbf{v}_i))

    where :math:`\mathcal{L}` is a specific loss function,
    default Mean Square Error:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    In this context :math:`\mathbf{u}_i` and :math:`\mathbf{v}_i` means that
    we are seeking to approximate multiple (discretised) functions given
    multiple (discretised) input functions.
    """

    def __init__(
        self,
        problem,
        model,
        extra_features=None,
        loss=torch.nn.MSELoss(),
        optimizer=torch.optim.Adam,
        optimizer_kwargs={"lr": 0.001},
        scheduler=ConstantLR,
        scheduler_kwargs={"factor": 1, "total_iters": 0},
    ):
        """
        :param AbstractProblem problem: The formualation of the problem.
        :param torch.nn.Module model: The neural network model to use.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default :class:`torch.nn.MSELoss`.
        :param torch.nn.Module extra_features: The additional input
            features to use as augmented input.
        :param torch.optim.Optimizer optimizer: The neural network optimizer to
            use; default is :class:`torch.optim.Adam`.
        :param dict optimizer_kwargs: Optimizer constructor keyword args.
        :param float lr: The learning rate; default is 0.001.
        :param torch.optim.LRScheduler scheduler: Learning
            rate scheduler.
        :param dict scheduler_kwargs: LR scheduler constructor keyword args.
        """
        super().__init__(
            models=[model],
            problem=problem,
            optimizers=[optimizer],
            optimizers_kwargs=[optimizer_kwargs],
            extra_features=extra_features,
        )

        # check consistency
        check_consistency(scheduler, LRScheduler, subclass=True)
        check_consistency(scheduler_kwargs, dict)
        check_consistency(loss, (LossInterface, _Loss), subclass=False)

        # assign variables
        self._scheduler = scheduler(self.optimizers[0], **scheduler_kwargs)
        self._loss = loss
        self._neural_net = self.models[0]

    def forward(self, x):
        """Forward pass implementation for the solver.

        :param torch.Tensor x: Input tensor.
        :return: Solver solution.
        :rtype: torch.Tensor
        """
        return self.neural_net(x)

    def configure_optimizers(self):
        """Optimizer configuration for the solver.

        :return: The optimizers and the schedulers
        :rtype: tuple(list, list)
        """
        return self.optimizers, [self.scheduler]

    def training_step(self, batch, batch_idx):
        """Solver training step.

        :param batch: The batch element in the dataloader.
        :type batch: tuple
        :param batch_idx: The batch index.
        :type batch_idx: int
        :return: The sum of the loss functions.
        :rtype: LabelTensor
        """

        condition_idx = batch["condition"]

        for condition_id in range(condition_idx.min(), condition_idx.max() + 1):

            condition_name = self._dataloader.condition_names[condition_id]
            condition = self.problem.conditions[condition_name]
            pts = batch["pts"]
            out = batch["output"]

            if condition_name not in self.problem.conditions:
                raise RuntimeError("Something wrong happened.")

            # for data driven mode
            if not hasattr(condition, "output_points"):
                raise NotImplementedError(
                    f"{type(self).__name__} works only in data-driven mode."
                )

            output_pts = out[condition_idx == condition_id]
            input_pts = pts[condition_idx == condition_id]

            loss = (
                self.loss_data(input_pts=input_pts, output_pts=output_pts)
                * condition.data_weight
            )
            loss = loss.as_subclass(torch.Tensor)

        self.log("mean_loss", float(loss), prog_bar=True, logger=True)
        return loss

    def loss_data(self, input_pts, output_pts):
        """
        The data loss for the Supervised solver. It computes the loss between
        the network output against the true solution. This function
        should not be override if not intentionally.

        :param LabelTensor input_tensor: The input to the neural networks.
        :param LabelTensor output_tensor: The true solution to compare the
            network solution.
        :return: The residual loss averaged on the input coordinates
        :rtype: torch.Tensor
        """
        return self.loss(self.forward(input_pts), output_pts)

    @property
    def scheduler(self):
        """
        Scheduler for training.
        """
        return self._scheduler

    @property
    def neural_net(self):
        """
        Neural network for training.
        """
        return self._neural_net

    @property
    def loss(self):
        """
        Loss for training.
        """
        return self._loss



""" Module for Averaging Neural Operator Layer class. """

import torch

from pina.utils import check_consistency
import pina.model as pm  # avoid circular import


class LowRankBlock(torch.nn.Module):
    r"""
    The PINA implementation of the inner layer of the Averaging Neural Operator.

    The operator layer performs an affine transformation where the convolution
    is approximated with a local average. Given the input function
    :math:`v(x)\in\mathbb{R}^{\rm{emb}}` the layer computes
    the operator update :math:`K(v)` as:

    .. math::
        K(v) = \sigma\left(Wv(x) + b + \sum_{i=1}^r \langle
        \psi^{(i)} , v(x) \rangle \phi^{(i)} \right)

    where:

    *   :math:`\mathbb{R}^{\rm{emb}}` is the embedding (hidden) size
        corresponding to the ``hidden_size`` object
    *   :math:`\sigma` is a non-linear activation, corresponding to the
        ``func`` object
    *   :math:`W\in\mathbb{R}^{\rm{emb}\times\rm{emb}}` is a tunable matrix.
    *   :math:`b\in\mathbb{R}^{\rm{emb}}` is a tunable bias.
    *   :math:`\psi^{(i)}\in\mathbb{R}^{\rm{emb}}` and
        :math:`\phi^{(i)}\in\mathbb{R}^{\rm{emb}}` are :math:`r` a low rank
        basis functions mapping.
    *   :math:`b\in\mathbb{R}^{\rm{emb}}` is a tunable bias.

    .. seealso::

        **Original reference**: Kovachki, N., Li, Z., Liu, B.,
        Azizzadenesheli, K., Bhattacharya, K., Stuart, A., & Anandkumar, A.
        (2023). *Neural operator: Learning maps between function
        spaces with applications to PDEs*. Journal of Machine Learning
        Research, 24(89), 1-97.

    """

    def __init__(
        self,
        input_dimensions,
        embedding_dimenion,
        rank,
        inner_size=20,
        n_layers=2,
        func=torch.nn.Tanh,
        bias=True,
    ):
        """
        :param int input_dimensions: The number of input components of the
            model.
            Expected tensor shape of the form :math:`(*, d)`, where *
            means any number of dimensions including none,
            and :math:`d` the ``input_dimensions``.
        :param int embedding_dimenion: Size of the embedding dimension of the
            field.
        :param int rank: The rank number of the basis approximation components
            of the model. Expected tensor shape of the form :math:`(*, 2d)`,
            where * means any number of dimensions including none,
            and :math:`2d` the ``rank`` for both basis functions.
        :param int inner_size: Number of neurons in the hidden layer(s) for the
            basis function network. Default is 20.
        :param int n_layers: Number of hidden layers. for the
            basis function network. Default is 2.
        :param func: The activation function to use for the
            basis function network. If a single
            :class:`torch.nn.Module` is passed, this is used as
            activation function after any layers, except the last one.
            If a list of Modules is passed,
            they are used as activation functions at any layers, in order.
        :param bool bias: If ``True`` the MLP will consider some bias for the
            basis function network.
        """
        super().__init__()

        # Assignment (check consistency inside FeedForward)
        self._basis = pm.FeedForward(
            input_dimensions=input_dimensions,
            output_dimensions=2 * rank * embedding_dimenion,
            inner_size=inner_size,
            n_layers=n_layers,
            func=func,
            bias=bias,
        )
        self._nn = torch.nn.Linear(embedding_dimenion, embedding_dimenion)

        check_consistency(rank, int)
        self._rank = rank
        self._func = func()

    def forward(self, x, coords):
        r"""
        Forward pass of the layer, it performs an affine transformation of
        the field, and a low rank approximation by
        doing a dot product of the basis
        :math:`\psi^{(i)}` with the filed vector :math:`v`, and use this
        coefficients to expand :math:`\phi^{(i)}` evaluated in the
        spatial input :math:`x`.

        :param torch.Tensor x: The input tensor for performing the
            computation. It expects a tensor :math:`B \times N \times D`,
            where :math:`B` is the batch_size, :math:`N` the number of points
            in the mesh, :math:`D` the dimension of the problem. In particular
            :math:`D` is the codomain of the function :math:`v`. For example
            a scalar function has :math:`D=1`, a 4-dimensional vector function
            :math:`D=4`.
        :param torch.Tensor coords: The coordinates in which the field is
            evaluated for performing the  computation. It expects a
            tensor :math:`B \times N \times d`,
            where :math:`B` is the batch_size, :math:`N` the number of points
            in the mesh, :math:`D` the dimension of the domain.
        :return: The output tensor obtained from Average Neural Operator Block.
        :rtype: torch.Tensor
        """
        # extract basis
        basis = self._basis(coords)
        # reshape [B, N, D, 2*rank]
        shape = list(basis.shape[:-1]) + [-1, 2 * self.rank]
        basis = basis.reshape(shape)
        # divide
        psi = basis[..., : self.rank]
        phi = basis[..., self.rank :]
        # compute dot product
        coeff = torch.einsum("...dr,...d->...r", psi, x)
        # expand the basis
        expansion = torch.einsum("...r,...dr->...d", coeff, phi)
        # apply linear layer and return
        return self._func(self._nn(x) + expansion)

    @property
    def rank(self):
        """
        The basis rank.
        """
        return self._rank



import torch
import pytest

from pina.adaptive_functions import (AdaptiveReLU, AdaptiveSigmoid, AdaptiveTanh,
                            AdaptiveSiLU, AdaptiveMish, AdaptiveELU,
                            AdaptiveCELU, AdaptiveGELU, AdaptiveSoftmin,
                            AdaptiveSoftmax, AdaptiveSIREN, AdaptiveExp)


adaptive_functions = (AdaptiveReLU, AdaptiveSigmoid, AdaptiveTanh,
                      AdaptiveSiLU, AdaptiveMish, AdaptiveELU,
                      AdaptiveCELU, AdaptiveGELU, AdaptiveSoftmin,
                      AdaptiveSoftmax, AdaptiveSIREN, AdaptiveExp)
x = torch.rand(10, requires_grad=True)

@pytest.mark.parametrize("Func", adaptive_functions)
def test_constructor(Func):
    if Func.__name__ == 'AdaptiveExp':
        # simple
        Func()
        # setting values
        af = Func(alpha=1., beta=2.)
        assert af.alpha.requires_grad
        assert af.beta.requires_grad
        assert af.alpha == 1.
        assert af.beta == 2.
    else:
        # simple
        Func()
        # setting values
        af = Func(alpha=1., beta=2., gamma=3.)
        assert af.alpha.requires_grad
        assert af.beta.requires_grad
        assert af.gamma.requires_grad
        assert af.alpha == 1.
        assert af.beta == 2.
        assert af.gamma == 3.

    # fixed variables
    af = Func(alpha=1., beta=2., fixed=['alpha'])
    assert af.alpha.requires_grad is False
    assert af.beta.requires_grad
    assert af.alpha == 1.
    assert af.beta == 2.

    with pytest.raises(TypeError):
        Func(alpha=1., beta=2., fixed=['delta'])

    with pytest.raises(ValueError):
        Func(alpha='s')
        Func(alpha=1)

@pytest.mark.parametrize("Func", adaptive_functions)      
def test_forward(Func):
    af = Func()
    af(x)

@pytest.mark.parametrize("Func", adaptive_functions)
def test_backward(Func):
    af = Func()
    y = af(x)
    y.mean().backward()


""" Module for GPINN """

import torch


from torch.optim.lr_scheduler import ConstantLR

from .pinn import PINN
from pina.operators import grad
from pina.problem import SpatialProblem


class GPINN(PINN):
    r"""
    Gradient Physics Informed Neural Network (GPINN) solver class.
    This class implements Gradient Physics Informed Neural
    Network solvers, using a user specified ``model`` to solve a specific
    ``problem``. It can be used for solving both forward and inverse problems.

    The Gradient Physics Informed Network aims to find
    the solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m`
    of the differential problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}

    minimizing the loss function

    .. math::
        \mathcal{L}_{\rm{problem}} =& \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{A}[\mathbf{u}](\mathbf{x}_i)) +
        \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{B}[\mathbf{u}](\mathbf{x}_i)) + \\
        &\frac{1}{N}\sum_{i=1}^N
        \nabla_{\mathbf{x}}\mathcal{L}(\mathcal{A}[\mathbf{u}](\mathbf{x}_i)) +
        \frac{1}{N}\sum_{i=1}^N
        \nabla_{\mathbf{x}}\mathcal{L}(\mathcal{B}[\mathbf{u}](\mathbf{x}_i))


    where :math:`\mathcal{L}` is a specific loss function, default Mean Square Error:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    .. seealso::

        **Original reference**: Yu, Jeremy, et al. "Gradient-enhanced
        physics-informed neural networks for forward and inverse
        PDE problems." Computer Methods in Applied Mechanics
        and Engineering 393 (2022): 114823.
        DOI: `10.1016 <https://doi.org/10.1016/j.cma.2022.114823>`_.

    .. note::
        This class can only work for problems inheriting
        from at least :class:`~pina.problem.spatial_problem.SpatialProblem`
        class.
    """

    def __init__(
        self,
        problem,
        model,
        extra_features=None,
        loss=torch.nn.MSELoss(),
        optimizer=torch.optim.Adam,
        optimizer_kwargs={"lr": 0.001},
        scheduler=ConstantLR,
        scheduler_kwargs={"factor": 1, "total_iters": 0},
    ):
        """
        :param AbstractProblem problem: The formulation of the problem. It must
            inherit from at least
            :class:`~pina.problem.spatial_problem.SpatialProblem` in order to
            compute the gradient of the loss.
        :param torch.nn.Module model: The neural network model to use.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default :class:`torch.nn.MSELoss`.
        :param torch.nn.Module extra_features: The additional input
            features to use as augmented input.
        :param torch.optim.Optimizer optimizer: The neural network optimizer to
            use; default is :class:`torch.optim.Adam`.
        :param dict optimizer_kwargs: Optimizer constructor keyword args.
        :param torch.optim.LRScheduler scheduler: Learning
            rate scheduler.
        :param dict scheduler_kwargs: LR scheduler constructor keyword args.
        """
        super().__init__(
            problem=problem,
            model=model,
            extra_features=extra_features,
            loss=loss,
            optimizer=optimizer,
            optimizer_kwargs=optimizer_kwargs,
            scheduler=scheduler,
            scheduler_kwargs=scheduler_kwargs,
        )
        if not isinstance(self.problem, SpatialProblem):
            raise ValueError(
                "Gradient PINN computes the gradient of the "
                "PINN loss with respect to the spatial "
                "coordinates, thus the PINA problem must be "
                "a SpatialProblem."
            )

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the GPINN solver based on given
        samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation
            representing the physics.
        :return: The physics loss calculated based on given
            samples and equation.
        :rtype: LabelTensor
        """
        # classical PINN loss
        residual = self.compute_residual(samples=samples, equation=equation)
        loss_value = self.loss(
            torch.zeros_like(residual, requires_grad=True), residual
        )
        self.store_log(loss_value=float(loss_value))
        # gradient PINN loss
        loss_value = loss_value.reshape(-1, 1)
        loss_value.labels = ["__LOSS"]
        loss_grad = grad(loss_value, samples, d=self.problem.spatial_variables)
        g_loss_phys = self.loss(
            torch.zeros_like(loss_grad, requires_grad=True), loss_grad
        )
        return loss_value + g_loss_phys



from pina.geometry import CartesianDomain
from pina import Condition, Plotter
from matplotlib.testing.decorators import image_comparison
import matplotlib.pyplot as plt
from pina.problem import SpatialProblem
from pina.equation import FixedValue

        
class FooProblem1D(SpatialProblem):

    # assign output/ spatial and temporal variables
    output_variables = ['u']
    spatial_domain = CartesianDomain({'x' : [-1, 1]})

    # problem condition statement
    conditions = {
        'D': Condition(location=CartesianDomain({'x': [-1, 1]}), equation=FixedValue(0.)),
    }

class FooProblem2D(SpatialProblem):

    # assign output/ spatial and temporal variables
    output_variables = ['u']
    spatial_domain = CartesianDomain({'x' : [-1, 1], 'y': [-1, 1]})

    # problem condition statement
    conditions = {
        'D': Condition(location=CartesianDomain({'x' : [-1, 1], 'y': [-1, 1]}), equation=FixedValue(0.)),
    }

class FooProblem3D(SpatialProblem):

    # assign output/ spatial and temporal variables
    output_variables = ['u']
    spatial_domain = CartesianDomain({'x' : [-1, 1], 'y': [-1, 1], 'z':[-1,1]})

    # problem condition statement
    conditions = {
        'D': Condition(location=CartesianDomain({'x' : [-1, 1], 'y': [-1, 1], 'z':[-1,1]}), equation=FixedValue(0.)),
    }



def test_constructor():
    Plotter()

def test_plot_samples_1d():
    problem = FooProblem1D()
    problem.discretise_domain(n=10, mode='grid', variables = 'x', locations=['D'])
    pl = Plotter()
    pl.plot_samples(problem=problem, filename='fig.png')
    import os
    os.remove('fig.png')

def test_plot_samples_2d():
    problem = FooProblem2D()
    problem.discretise_domain(n=10, mode='grid', variables = ['x', 'y'], locations=['D'])
    pl = Plotter()
    pl.plot_samples(problem=problem, filename='fig.png')
    import os
    os.remove('fig.png')

def test_plot_samples_3d():
    problem = FooProblem3D()
    problem.discretise_domain(n=10, mode='grid', variables = ['x', 'y', 'z'], locations=['D'])
    pl = Plotter()
    pl.plot_samples(problem=problem, filename='fig.png')
    import os
    os.remove('fig.png')


""" Module for CompetitivePINN """

import torch
import copy

try:
    from torch.optim.lr_scheduler import LRScheduler  # torch >= 2.0
except ImportError:
    from torch.optim.lr_scheduler import (
        _LRScheduler as LRScheduler,
    )  # torch < 2.0

from torch.optim.lr_scheduler import ConstantLR

from .basepinn import PINNInterface
from pina.utils import check_consistency
from pina.problem import InverseProblem


class CompetitivePINN(PINNInterface):
    r"""
    Competitive Physics Informed Neural Network (PINN) solver class.
    This class implements Competitive Physics Informed Neural
    Network solvers, using a user specified ``model`` to solve a specific
    ``problem``. It can be used for solving both forward and inverse problems.

    The Competitive Physics Informed Network aims to find
    the solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m`
    of the differential problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}

    with a minimization (on ``model`` parameters) maximation (
    on ``discriminator`` parameters) of the loss function

    .. math::
        \mathcal{L}_{\rm{problem}} = \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(D(\mathbf{x}_i)\mathcal{A}[\mathbf{u}](\mathbf{x}_i))+
        \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(D(\mathbf{x}_i)\mathcal{B}[\mathbf{u}](\mathbf{x}_i))

    where :math:`D` is the discriminator network, which tries to find the points
    where the network performs worst, and :math:`\mathcal{L}` is a specific loss
    function, default Mean Square Error:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.

    .. seealso::

        **Original reference**: Zeng, Qi, et al.
        "Competitive physics informed networks." International Conference on
        Learning Representations, ICLR 2022
        `OpenReview Preprint <https://openreview.net/forum?id=z9SIj-IM7tn>`_.

    .. warning::
        This solver does not currently support the possibility to pass
        ``extra_feature``.
    """

    def __init__(
        self,
        problem,
        model,
        discriminator=None,
        loss=torch.nn.MSELoss(),
        optimizer_model=torch.optim.Adam,
        optimizer_model_kwargs={"lr": 0.001},
        optimizer_discriminator=torch.optim.Adam,
        optimizer_discriminator_kwargs={"lr": 0.001},
        scheduler_model=ConstantLR,
        scheduler_model_kwargs={"factor": 1, "total_iters": 0},
        scheduler_discriminator=ConstantLR,
        scheduler_discriminator_kwargs={"factor": 1, "total_iters": 0},
    ):
        """
        :param AbstractProblem problem: The formualation of the problem.
        :param torch.nn.Module model: The neural network model to use
            for the model.
        :param torch.nn.Module discriminator: The neural network model to use
            for the discriminator. If ``None``, the discriminator network will
            have the same architecture as the model network.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default :class:`torch.nn.MSELoss`.
        :param torch.optim.Optimizer optimizer_model: The neural
            network optimizer to use for the model network
            , default is `torch.optim.Adam`.
        :param dict optimizer_model_kwargs: Optimizer constructor keyword
            args. for the model.
        :param torch.optim.Optimizer optimizer_discriminator: The neural
            network optimizer to use for the discriminator network
            , default is `torch.optim.Adam`.
        :param dict optimizer_discriminator_kwargs: Optimizer constructor
            keyword args. for the discriminator.
        :param torch.optim.LRScheduler scheduler_model: Learning
            rate scheduler for the model.
        :param dict scheduler_model_kwargs: LR scheduler constructor
            keyword args.
        :param torch.optim.LRScheduler scheduler_discriminator: Learning
            rate scheduler for the discriminator.
        """
        if discriminator is None:
            discriminator = copy.deepcopy(model)

        super().__init__(
            models=[model, discriminator],
            problem=problem,
            optimizers=[optimizer_model, optimizer_discriminator],
            optimizers_kwargs=[
                optimizer_model_kwargs,
                optimizer_discriminator_kwargs,
            ],
            extra_features=None,  # CompetitivePINN doesn't take extra features
            loss=loss,
        )

        # set automatic optimization for GANs
        self.automatic_optimization = False

        # check consistency
        check_consistency(scheduler_model, LRScheduler, subclass=True)
        check_consistency(scheduler_model_kwargs, dict)
        check_consistency(scheduler_discriminator, LRScheduler, subclass=True)
        check_consistency(scheduler_discriminator_kwargs, dict)

        # assign schedulers
        self._schedulers = [
            scheduler_model(self.optimizers[0], **scheduler_model_kwargs),
            scheduler_discriminator(
                self.optimizers[1], **scheduler_discriminator_kwargs
            ),
        ]

        self._model = self.models[0]
        self._discriminator = self.models[1]

    def forward(self, x):
        r"""
        Forward pass implementation for the PINN solver. It returns the function
        evaluation :math:`\mathbf{u}(\mathbf{x})` at the control points
        :math:`\mathbf{x}`.

        :param LabelTensor x: Input tensor for the PINN solver. It expects
            a tensor :math:`N \times D`, where :math:`N` the number of points
            in the mesh, :math:`D` the dimension of the problem,
        :return: PINN solution evaluated at contro points.
        :rtype: LabelTensor
        """
        return self.neural_net(x)

    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the Competitive PINN solver based on given
        samples and equation.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation
            representing the physics.
        :return: The physics loss calculated based on given
            samples and equation.
        :rtype: LabelTensor
        """
        # train one step of the model
        with torch.no_grad():
            discriminator_bets = self.discriminator(samples)
        loss_val = self._train_model(samples, equation, discriminator_bets)
        self.store_log(loss_value=float(loss_val))
        # detaching samples from the computational graph to erase it and setting
        # the gradient to true to create a new computational graph.
        # In alternative set `retain_graph=True`.
        samples = samples.detach()
        samples.requires_grad = True
        # train one step of discriminator
        discriminator_bets = self.discriminator(samples)
        self._train_discriminator(samples, equation, discriminator_bets)
        return loss_val

    def loss_data(self, input_tensor, output_tensor):
        """
        The data loss for the PINN solver. It computes the loss between the
        network output against the true solution.

        :param LabelTensor input_tensor: The input to the neural networks.
        :param LabelTensor output_tensor: The true solution to compare the
            network solution.
        :return: The computed data loss.
        :rtype: torch.Tensor
        """
        self.optimizer_model.zero_grad()
        loss_val = (
            super()
            .loss_data(input_tensor, output_tensor)
            .as_subclass(torch.Tensor)
        )
        loss_val.backward()
        self.optimizer_model.step()
        return loss_val

    def configure_optimizers(self):
        """
        Optimizer configuration for the Competitive PINN solver.

        :return: The optimizers and the schedulers
        :rtype: tuple(list, list)
        """
        # if the problem is an InverseProblem, add the unknown parameters
        # to the parameters that the optimizer needs to optimize
        if isinstance(self.problem, InverseProblem):
            self.optimizer_model.add_param_group(
                {
                    "params": [
                        self._params[var]
                        for var in self.problem.unknown_variables
                    ]
                }
            )
        return self.optimizers, self._schedulers

    def on_train_batch_end(self, outputs, batch, batch_idx):
        """
        This method is called at the end of each training batch, and ovverides
        the PytorchLightining implementation for logging the checkpoints.

        :param torch.Tensor outputs: The output from the model for the
            current batch.
        :param tuple batch: The current batch of data.
        :param int batch_idx: The index of the current batch.
        :return: Whatever is returned by the parent
            method ``on_train_batch_end``.
        :rtype: Any
        """
        # increase by one the counter of optimization to save loggers
        self.trainer.fit_loop.epoch_loop.manual_optimization.optim_step_progress.total.completed += (
            1
        )
        return super().on_train_batch_end(outputs, batch, batch_idx)

    def _train_discriminator(self, samples, equation, discriminator_bets):
        """
        Trains the discriminator network of the Competitive PINN.

        :param LabelTensor samples: Input samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation representing
            the physics.
        :param Tensor discriminator_bets: Predictions made by the discriminator
            network.
        """
        # manual optimization
        self.optimizer_discriminator.zero_grad()
        # compute residual, we detach because the weights of the generator
        # model are fixed
        residual = self.compute_residual(
            samples=samples, equation=equation
        ).detach()
        # compute competitive residual, the minus is because we maximise
        competitive_residual = residual * discriminator_bets
        loss_val = -self.loss(
            torch.zeros_like(competitive_residual, requires_grad=True),
            competitive_residual,
        ).as_subclass(torch.Tensor)
        # backprop
        self.manual_backward(loss_val)
        self.optimizer_discriminator.step()
        return

    def _train_model(self, samples, equation, discriminator_bets):
        """
        Trains the model network of the Competitive PINN.

        :param LabelTensor samples: Input samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation representing
            the physics.
        :param Tensor discriminator_bets: Predictions made by the discriminator.
            network.
        :return: The computed data loss.
        :rtype: torch.Tensor
        """
        # manual optimization
        self.optimizer_model.zero_grad()
        # compute residual (detached for discriminator) and log
        residual = self.compute_residual(samples=samples, equation=equation)
        # store logging
        with torch.no_grad():
            loss_residual = self.loss(torch.zeros_like(residual), residual)
        # compute competitive residual, discriminator_bets are detached becase
        # we optimize only the generator model
        competitive_residual = residual * discriminator_bets.detach()
        loss_val = self.loss(
            torch.zeros_like(competitive_residual, requires_grad=True),
            competitive_residual,
        ).as_subclass(torch.Tensor)
        # backprop
        self.manual_backward(loss_val)
        self.optimizer_model.step()
        return loss_residual

    @property
    def neural_net(self):
        """
        Returns the neural network model.

        :return: The neural network model.
        :rtype: torch.nn.Module
        """
        return self._model

    @property
    def discriminator(self):
        """
        Returns the discriminator model (if applicable).

        :return: The discriminator model.
        :rtype: torch.nn.Module
        """
        return self._discriminator

    @property
    def optimizer_model(self):
        """
        Returns the optimizer associated with the neural network model.

        :return: The optimizer for the neural network model.
        :rtype: torch.optim.Optimizer
        """
        return self.optimizers[0]

    @property
    def optimizer_discriminator(self):
        """
        Returns the optimizer associated with the discriminator (if applicable).

        :return: The optimizer for the discriminator.
        :rtype: torch.optim.Optimizer
        """
        return self.optimizers[1]

    @property
    def scheduler_model(self):
        """
        Returns the scheduler associated with the neural network model.

        :return: The scheduler for the neural network model.
        :rtype: torch.optim.lr_scheduler._LRScheduler
        """
        return self._schedulers[0]

    @property
    def scheduler_discriminator(self):
        """
        Returns the scheduler associated with the discriminator (if applicable).

        :return: The scheduler for the discriminator.
        :rtype: torch.optim.lr_scheduler._LRScheduler
        """
        return self._schedulers[1]



""" Periodic Boundary Embedding modulus. """

import torch
from pina.utils import check_consistency


class PeriodicBoundaryEmbedding(torch.nn.Module):
    r"""
    Imposing hard constraint periodic boundary conditions by embedding the
    input.

    A periodic function :math:`u:\mathbb{R}^{\rm{in}}
    \rightarrow\mathbb{R}^{\rm{out}}` periodic in the spatial
    coordinates :math:`\mathbf{x}` with periods :math:`\mathbf{L}` is such that:

    .. math::
        u(\mathbf{x})  = u(\mathbf{x} + n \mathbf{L})\;\;
        \forall n\in\mathbb{N}.

    The :meth:`PeriodicBoundaryEmbedding` augments the input such that the periodic conditons
    is guarantee. The input is augmented by the following formula:

    .. math::
        \mathbf{x} \rightarrow \tilde{\mathbf{x}} = \left[1,
        \cos\left(\frac{2\pi}{L_1} x_1 \right),
        \sin\left(\frac{2\pi}{L_1}x_1\right), \cdots,
        \cos\left(\frac{2\pi}{L_{\rm{in}}}x_{\rm{in}}\right),
        \sin\left(\frac{2\pi}{L_{\rm{in}}}x_{\rm{in}}\right)\right],

    where :math:`\text{dim}(\tilde{\mathbf{x}}) = 3\text{dim}(\mathbf{x})`.

    .. seealso::
        **Original reference**:
            1.  Dong, Suchuan, and Naxian Ni (2021). *A method for representing
                periodic functions and enforcing exactly periodic boundary
                conditions with deep neural networks*. Journal of Computational
                Physics 435, 110242.
                DOI: `10.1016/j.jcp.2021.110242.
                <https://doi.org/10.1016/j.jcp.2021.110242>`_
            2.  Wang, S., Sankaran, S., Wang, H., & Perdikaris, P. (2023). *An
                expert's guide to training physics-informed neural networks*.
                DOI: `arXiv preprint arXiv:2308.0846.
                <https://arxiv.org/abs/2308.08468>`_
    .. warning::
        The embedding is a truncated fourier expansion, and only ensures
        function PBC and not for its derivatives. Ensuring approximate
        periodicity in
        the derivatives of :math:`u` can be done, and extensive
        tests have shown (also in the reference papers) that this implementation
        can correctly compute the PBC on the derivatives up to the order
        :math:`\sim 2,3`, while it is not guarantee the periodicity for
        :math:`>3`. The PINA code is tested only for function PBC and not for
        its derivatives.
    """

    def __init__(self, input_dimension, periods, output_dimension=None):
        """
        :param int input_dimension: The dimension of the input tensor, it can
            be checked with `tensor.ndim` method.
        :param float | int | dict periods: The periodicity in each dimension for
            the input data. If ``float`` or ``int`` is passed,
            the period is assumed constant for all the dimensions of the data.
            If a ``dict`` is passed the `dict.values` represent periods,
            while the ``dict.keys`` represent the dimension where the
            periodicity is applied. The `dict.keys` can either be `int`
            if working with ``torch.Tensor`` or ``str`` if
            working with ``LabelTensor``.
        :param int output_dimension: The dimension of the output after the
            fourier embedding. If not ``None`` a ``torch.nn.Linear`` layer
            is applied to the fourier embedding output to match the desired
            dimensionality, default ``None``.
        """
        super().__init__()

        # check input consistency
        check_consistency(periods, (float, int, dict))
        check_consistency(input_dimension, int)
        if output_dimension is not None:
            check_consistency(output_dimension, int)
            self._layer = torch.nn.Linear(input_dimension * 3, output_dimension)
        else:
            self._layer = torch.nn.Identity()

        # checks on the periods
        if isinstance(periods, dict):
            if not all(
                isinstance(dim, (str, int)) and isinstance(period, (float, int))
                for dim, period in periods.items()
            ):
                raise TypeError(
                    "In dictionary periods, keys must be integers"
                    " or strings, and values must be float or int."
                )
            self._period = periods
        else:
            self._period = {k: periods for k in range(input_dimension)}

    def forward(self, x):
        """
        Forward pass to compute the periodic boundary conditions embedding.

        :param torch.Tensor x: Input tensor.
        :return: Fourier embeddings of the input.
        :rtype: torch.Tensor
        """
        omega = torch.stack(
            [
                torch.pi * 2.0 / torch.tensor([val], device=x.device)
                for val in self._period.values()
            ],
            dim=-1,
        )
        x = self._get_vars(x, list(self._period.keys()))
        return self._layer(
            torch.cat(
                [
                    torch.ones_like(x),
                    torch.cos(omega * x),
                    torch.sin(omega * x),
                ],
                dim=-1,
            )
        )

    def _get_vars(self, x, indeces):
        """
        Get variables from input tensor ordered by specific indeces.

        :param torch.Tensor x: The input tensor to extract.
        :param list[int] | list[str] indeces: List of indeces to extract.
        :return: The extracted tensor given the indeces.
        :rtype: torch.Tensor
        """
        if isinstance(indeces[0], str):
            try:
                return x.extract(indeces)
            except AttributeError:
                raise RuntimeError(
                    "Not possible to extract input variables from tensor."
                    " Ensure that the passed tensor is a LabelTensor or"
                    " pass list of integers to extract variables. For"
                    " more information refer to warning in the documentation."
                )
        elif isinstance(indeces[0], int):
            return x[..., indeces]
        else:
            raise RuntimeError(
                "Not able to extract right indeces for tensor."
                " For more information refer to warning in the documentation."
            )

    @property
    def period(self):
        """
        The period of the periodic function to approximate.
        """
        return self._period



"""Module Averaging Neural Operator."""

import torch
from torch import nn, concatenate
from .layers import AVNOBlock
from .base_no import KernelNeuralOperator
from pina.utils import check_consistency


class AveragingNeuralOperator(KernelNeuralOperator):
    """
    Implementation of Averaging Neural Operator.

    Averaging Neural Operator is a general architecture for
    learning Operators. Unlike traditional machine learning methods
    AveragingNeuralOperator is designed to map entire functions
    to other functions. It can be trained with Supervised learning strategies.
    AveragingNeuralOperator does convolution by performing a field average.

    .. seealso::

        **Original reference**: Lanthaler S. Li, Z., Kovachki,
        Stuart, A. (2020). *The Nonlocal Neural Operator:
        Universal Approximation*.
        DOI: `arXiv preprint arXiv:2304.13221.
        <https://arxiv.org/abs/2304.13221>`_
    """

    def __init__(
        self,
        lifting_net,
        projecting_net,
        field_indices,
        coordinates_indices,
        n_layers=4,
        func=nn.GELU,
    ):
        """
        :param torch.nn.Module lifting_net: The neural network for lifting
            the input. It must take as input the input field and the coordinates
            at which the input field is avaluated. The output of the lifting
            net is chosen as embedding dimension of the problem
        :param torch.nn.Module projecting_net: The neural network for
            projecting the output. It must take as input the embedding dimension
            (output of the ``lifting_net``) plus the dimension
            of the coordinates.
        :param list[str] field_indices: the label of the fields
            in the input tensor.
        :param list[str] coordinates_indices: the label of the
            coordinates in the input tensor.
        :param int n_layers: number of hidden layers. Default is 4.
        :param torch.nn.Module func: the activation function to use,
            default to torch.nn.GELU.
        """

        # check consistency
        check_consistency(field_indices, str)
        check_consistency(coordinates_indices, str)
        check_consistency(n_layers, int)
        check_consistency(func, nn.Module, subclass=True)

        # check hidden dimensions match
        input_lifting_net = next(lifting_net.parameters()).size()[-1]
        output_lifting_net = lifting_net(
            torch.rand(size=next(lifting_net.parameters()).size())
        ).shape[-1]
        projecting_net_input = next(projecting_net.parameters()).size()[-1]

        if len(field_indices) + len(coordinates_indices) != input_lifting_net:
            raise ValueError(
                "The lifting_net must take as input the "
                "coordinates vector and the field vector."
            )

        if (
            output_lifting_net + len(coordinates_indices)
            != projecting_net_input
        ):
            raise ValueError(
                "The projecting_net input must be equal to"
                "the embedding dimension (which is the output) "
                "of the lifting_net plus the dimension of the "
                "coordinates, i.e. len(coordinates_indices)."
            )

        # assign
        self.coordinates_indices = coordinates_indices
        self.field_indices = field_indices
        integral_net = nn.Sequential(
            *[AVNOBlock(output_lifting_net, func) for _ in range(n_layers)]
        )
        super().__init__(lifting_net, integral_net, projecting_net)

    def forward(self, x):
        r"""
        Forward computation for Averaging Neural Operator. It performs a
        lifting of the input by the ``lifting_net``. Then different layers
        of Averaging Neural Operator Blocks are applied.
        Finally the output is projected to the final dimensionality
        by the ``projecting_net``.

        :param torch.Tensor x: The input tensor for fourier block,
            depending on ``dimension`` in the initialization. It expects
            a tensor :math:`B \times N \times D`,
            where :math:`B` is the batch_size, :math:`N` the number of points
            in the mesh, :math:`D` the dimension of the problem, i.e. the sum
            of ``len(coordinates_indices)+len(field_indices)``.
        :return: The output tensor obtained from Average Neural Operator.
        :rtype: torch.Tensor
        """
        points_tmp = x.extract(self.coordinates_indices)
        new_batch = x.extract(self.field_indices)
        new_batch = concatenate((new_batch, points_tmp), dim=-1)
        new_batch = self._lifting_operator(new_batch)
        new_batch = self._integral_kernels(new_batch)
        new_batch = concatenate((new_batch, points_tmp), dim=-1)
        new_batch = self._projection_operator(new_batch)
        return new_batch



"""Module for Base Continuous Convolution class."""

from abc import ABCMeta, abstractmethod
import torch
from .stride import Stride
from .utils_convolution import optimizing


class PODBlock(torch.nn.Module):
    """
    POD layer: it projects the input field on the proper orthogonal
    decomposition basis.  It needs to be fitted to the data before being used
    with the method :meth:`fit`, which invokes the singular value decomposition.
    The layer is not trainable.

    .. note::
        All the POD modes are stored in memory, avoiding to recompute them when the rank changes but increasing the memory usage.
    """

    def __init__(self, rank, scale_coefficients=True):
        """
        Build the POD layer with the given rank.

        :param int rank: The rank of the POD layer.
        :param bool scale_coefficients: If True, the coefficients are scaled
            after the projection to have zero mean and unit variance.
        """
        super().__init__()
        self.__scale_coefficients = scale_coefficients
        self._basis = None
        self._scaler = None
        self._rank = rank

    @property
    def rank(self):
        """
        The rank of the POD layer.

        :rtype: int
        """
        return self._rank

    @rank.setter
    def rank(self, value):
        if value < 1 or not isinstance(value, int):
            raise ValueError("The rank must be positive integer")

        self._rank = value

    @property
    def basis(self):
        """
        The POD basis. It is a matrix whose columns are the first `self.rank` POD modes.

        :rtype: torch.Tensor
        """
        if self._basis is None:
            return None

        return self._basis[: self.rank]

    @property
    def scaler(self):
        """
        The scaler. It is a dictionary with the keys `'mean'` and `'std'` that
        store the mean and the standard deviation of the coefficients.

        :rtype: dict
        """
        if self._scaler is None:
            return

        return {
            "mean": self._scaler["mean"][: self.rank],
            "std": self._scaler["std"][: self.rank],
        }

    @property
    def scale_coefficients(self):
        """
        If True, the coefficients are scaled after the projection to have zero
        mean and unit variance.

        :rtype: bool
        """
        return self.__scale_coefficients

    def fit(self, X):
        """
        Set the POD basis by performing the singular value decomposition of the
        given tensor. If `self.scale_coefficients` is True, the coefficients
        are scaled after the projection to have zero mean and unit variance.

        :param torch.Tensor X: The tensor to be reduced.
        """
        self._fit_pod(X)

        if self.__scale_coefficients:
            self._fit_scaler(torch.matmul(self._basis, X.T))

    def _fit_scaler(self, coeffs):
        """
        Private merhod that computes the mean and the standard deviation of the
        given coefficients, allowing to scale them to have zero mean and unit
        variance. Mean and standard deviation are stored in the private member
        `_scaler`.

        :param torch.Tensor coeffs: The coefficients to be scaled.
        """
        self._scaler = {
            "std": torch.std(coeffs, dim=1),
            "mean": torch.mean(coeffs, dim=1),
        }

    def _fit_pod(self, X):
        """
        Private method that computes the POD basis of the given tensor and stores it in the private member `_basis`.

        :param torch.Tensor X: The tensor to be reduced.
        """
        if X.device.type == "mps":  #  svd_lowrank not arailable for mps
            self._basis = torch.svd(X.T)[0].T
        else:
            self._basis = torch.svd_lowrank(X.T, q=X.shape[0])[0].T

    def forward(self, X):
        """
        The forward pass of the POD layer. By default it executes the
        :meth:`reduce` method, reducing the input tensor to its POD
        representation. The POD layer needs to be fitted before being used.

        :param torch.Tensor X: The input tensor to be reduced.
        :return: The reduced tensor.
        :rtype: torch.Tensor
        """
        return self.reduce(X)

    def reduce(self, X):
        """
        Reduce the input tensor to its POD representation. The POD layer needs
        to be fitted before being used.

        :param torch.Tensor X: The input tensor to be reduced.
        :return: The reduced tensor.
        :rtype: torch.Tensor
        """
        if self._basis is None:
            raise RuntimeError(
                "The POD layer needs to be fitted before being used."
            )

        coeff = torch.matmul(self.basis, X.T)
        if coeff.ndim == 1:
            coeff = coeff.unsqueeze(1)

        coeff = coeff.T
        if self.__scale_coefficients:
            coeff = (coeff - self.scaler["mean"]) / self.scaler["std"]

        return coeff

    def expand(self, coeff):
        """
        Expand the given coefficients to the original space. The POD layer needs
        to be fitted before being used.

        :param torch.Tensor coeff: The coefficients to be expanded.
        :return: The expanded tensor.
        :rtype: torch.Tensor
        """
        if self._basis is None:
            raise RuntimeError(
                "The POD layer needs to be trained before being used."
            )

        if self.__scale_coefficients:
            coeff = coeff * self.scaler["std"] + self.scaler["mean"]
        predicted = torch.matmul(self.basis.T, coeff.T).T

        if predicted.ndim == 1:
            predicted = predicted.unsqueeze(0)

        return predicted



"""
Kernel Neural Operator Module.
"""

import torch
from pina.utils import check_consistency


class KernelNeuralOperator(torch.nn.Module):
    r"""
    Base class for composing Neural Operators with integral kernels.

    This is a base class for composing neural operators with multiple
    integral kernels. All neural operator models defined in PINA inherit
    from this class. The structure is inspired by the work of Kovachki, N.
    et al. see Figure 2 of the reference for extra details. The Neural
    Operators inheriting from this class can be written as:

    .. math::
        G_\theta  := P \circ K_m \circ \cdot \circ K_1 \circ L

    where:

    *   :math:`G_\theta: \mathcal{A}\subset \mathbb{R}^{\rm{in}} \rightarrow
        \mathcal{D}\subset \mathbb{R}^{\rm{out}}` is the neural operator
        approximation of the unknown real operator :math:`G`, that is
        :math:`G \approx G_\theta`
    *   :math:`L: \mathcal{A}\subset \mathbb{R}^{\rm{in}} \rightarrow
        \mathbb{R}^{\rm{emb}}` is a lifting operator mapping the input
        from its domain :math:`\mathcal{A}\subset \mathbb{R}^{\rm{in}}`
        to its embedding dimension :math:`\mathbb{R}^{\rm{emb}}`
    *   :math:`\{K_i : \mathbb{R}^{\rm{emb}} \rightarrow
        \mathbb{R}^{\rm{emb}} \}_{i=1}^m` are :math:`m` integral kernels
        mapping each hidden representation to the next one.
    *   :math:`P : \mathbb{R}^{\rm{emb}} \rightarrow  \mathcal{D}\subset
        \mathbb{R}^{\rm{out}}` is a projection operator mapping the hidden
        representation to the output function.

    .. seealso::

        **Original reference**: Kovachki, N., Li, Z., Liu, B.,
        Azizzadenesheli, K., Bhattacharya, K., Stuart, A., & Anandkumar, A.
        (2023). *Neural operator: Learning maps between function
        spaces with applications to PDEs*. Journal of Machine Learning
        Research, 24(89), 1-97.
    """

    def __init__(self, lifting_operator, integral_kernels, projection_operator):
        """
        :param torch.nn.Module lifting_operator: The lifting operator
            mapping the input to its hidden dimension.
        :param torch.nn.Module integral_kernels: List of integral kernels
            mapping each hidden representation to the next one.
        :param torch.nn.Module projection_operator: The projection operator
            mapping the hidden representation to the output function.
        """

        super().__init__()

        self._lifting_operator = lifting_operator
        self._integral_kernels = integral_kernels
        self._projection_operator = projection_operator

    @property
    def lifting_operator(self):
        """
        The lifting operator property.
        """
        return self._lifting_operator

    @lifting_operator.setter
    def lifting_operator(self, value):
        """
        The lifting operator setter

        :param torch.nn.Module value: The lifting operator torch module.
        """
        check_consistency(value, torch.nn.Module)
        self._lifting_operator = value

    @property
    def projection_operator(self):
        """
        The projection operator property.
        """
        return self._projection_operator

    @projection_operator.setter
    def projection_operator(self, value):
        """
        The projection operator setter

        :param torch.nn.Module value: The projection operator torch module.
        """
        check_consistency(value, torch.nn.Module)
        self._projection_operator = value

    @property
    def integral_kernels(self):
        """
        The integral kernels operator property.
        """
        return self._integral_kernels

    @integral_kernels.setter
    def integral_kernels(self, value):
        """
        The integral kernels operator setter

        :param torch.nn.Module value: The integral kernels operator torch
            module.
        """
        check_consistency(value, torch.nn.Module)
        self._integral_kernels = value

    def forward(self, x):
        r"""
        Forward computation for Base Neural Operator. It performs a
        lifting of the input by the ``lifting_operator``.
        Then different layers integral kernels are applied using
        ``integral_kernels``. Finally the output is projected
        to the final dimensionality by the ``projection_operator``.

        :param torch.Tensor x: The input tensor for performing the
            computation. It expects a tensor :math:`B \times N \times D`,
            where :math:`B` is the batch_size, :math:`N` the number of points
            in the mesh, :math:`D` the dimension of the problem. In particular
            :math:`D` is the number of spatial/paramtric/temporal variables
            plus the field variables. For example for 2D problems with 2
            output\ variables :math:`D=4`.
        :return: The output tensor obtained from the NO.
        :rtype: torch.Tensor
        """
        x = self.lifting_operator(x)
        x = self.integral_kernels(x)
        x = self.projection_operator(x)
        return x



""" Module for ReducedOrderModelSolver """

import torch

from pina.solvers import SupervisedSolver


class ReducedOrderModelSolver(SupervisedSolver):
    r"""
    ReducedOrderModelSolver solver class. This class implements a
    Reduced Order Model solver, using user specified ``reduction_network`` and
    ``interpolation_network`` to solve a specific ``problem``.

    The  Reduced Order Model approach aims to find
    the solution :math:`\mathbf{u}:\Omega\rightarrow\mathbb{R}^m`
    of the differential problem:

    .. math::

        \begin{cases}
        \mathcal{A}[\mathbf{u}(\mu)](\mathbf{x})=0\quad,\mathbf{x}\in\Omega\\
        \mathcal{B}[\mathbf{u}(\mu)](\mathbf{x})=0\quad,
        \mathbf{x}\in\partial\Omega
        \end{cases}

    This is done by using two neural networks. The ``reduction_network``, which
    contains an encoder :math:`\mathcal{E}_{\rm{net}}`, a decoder
    :math:`\mathcal{D}_{\rm{net}}`; and an ``interpolation_network``
    :math:`\mathcal{I}_{\rm{net}}`. The input is assumed to be discretised in
    the spatial dimensions.

    The following loss function is minimized during training

    .. math::
        \mathcal{L}_{\rm{problem}} = \frac{1}{N}\sum_{i=1}^N
        \mathcal{L}(\mathcal{E}_{\rm{net}}[\mathbf{u}(\mu_i)] -
        \mathcal{I}_{\rm{net}}[\mu_i]) + 
        \mathcal{L}(
            \mathcal{D}_{\rm{net}}[\mathcal{E}_{\rm{net}}[\mathbf{u}(\mu_i)]] -
            \mathbf{u}(\mu_i))

    where :math:`\mathcal{L}` is a specific loss function, default Mean Square Error:

    .. math::
        \mathcal{L}(v) = \| v \|^2_2.


    .. seealso::

        **Original reference**: Hesthaven, Jan S., and Stefano Ubbiali.
        "Non-intrusive reduced order modeling of nonlinear problems
        using neural networks." Journal of Computational
        Physics 363 (2018): 55-78.
        DOI `10.1016/j.jcp.2018.02.037
        <https://doi.org/10.1016/j.jcp.2018.02.037>`_.
        
    .. note::
        The specified ``reduction_network`` must contain two methods,
        namely ``encode`` for input encoding and ``decode`` for decoding the
        former result. The ``interpolation_network`` network ``forward`` output
        represents the interpolation of the latent space obtain with
        ``reduction_network.encode``.

    .. note::
        This solver uses the end-to-end training strategy, i.e. the
        ``reduction_network`` and ``interpolation_network`` are trained
        simultaneously. For reference on this trainig strategy look at:
        Pichi, Federico, Beatriz Moya, and Jan S. Hesthaven. 
        "A graph convolutional autoencoder approach to model order reduction
        for parametrized PDEs." Journal of
        Computational Physics 501 (2024): 112762.
        DOI 
        `10.1016/j.jcp.2024.112762 <https://doi.org/10.1016/
        j.jcp.2024.112762>`_.

    .. warning::
        This solver works only for data-driven model. Hence in the ``problem``
        definition the codition must only contain ``input_points``
        (e.g. coefficient parameters, time parameters), and ``output_points``.

    .. warning::
        This solver does not currently support the possibility to pass
        ``extra_feature``.
    """

    def __init__(
        self,
        problem,
        reduction_network,
        interpolation_network,
        loss=torch.nn.MSELoss(),
        optimizer=torch.optim.Adam,
        optimizer_kwargs={"lr": 0.001},
        scheduler=torch.optim.lr_scheduler.ConstantLR,
        scheduler_kwargs={"factor": 1, "total_iters": 0},
    ):
        """
        :param AbstractProblem problem: The formualation of the problem.
        :param torch.nn.Module reduction_network: The reduction network used
            for reducing the input space. It must contain two methods,
            namely ``encode`` for input encoding and ``decode`` for decoding the
            former result.
        :param torch.nn.Module interpolation_network: The interpolation network
            for interpolating the control parameters to latent space obtain by
            the ``reduction_network`` encoding.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default :class:`torch.nn.MSELoss`.
        :param torch.nn.Module extra_features: The additional input
            features to use as augmented input.
        :param torch.optim.Optimizer optimizer: The neural network optimizer to
            use; default is :class:`torch.optim.Adam`.
        :param dict optimizer_kwargs: Optimizer constructor keyword args.
        :param float lr: The learning rate; default is 0.001.
        :param torch.optim.LRScheduler scheduler: Learning
            rate scheduler.
        :param dict scheduler_kwargs: LR scheduler constructor keyword args.
        """
        model = torch.nn.ModuleDict(
            {
                "reduction_network": reduction_network,
                "interpolation_network": interpolation_network,
            }
        )

        super().__init__(
            model=model,
            problem=problem,
            loss=loss,
            optimizer=optimizer,
            optimizer_kwargs=optimizer_kwargs,
            scheduler=scheduler,
            scheduler_kwargs=scheduler_kwargs,
        )

        # assert reduction object contains encode/ decode
        if not hasattr(self.neural_net["reduction_network"], "encode"):
            raise SyntaxError(
                "reduction_network must have encode method. "
                "The encode method should return a lower "
                "dimensional representation of the input."
            )
        if not hasattr(self.neural_net["reduction_network"], "decode"):
            raise SyntaxError(
                "reduction_network must have decode method. "
                "The decode method should return a high "
                "dimensional representation of the encoding."
            )

    def forward(self, x):
        """
        Forward pass implementation for the solver. It finds the encoder
        representation by calling ``interpolation_network.forward`` on the
        input, and maps this representation to output space by calling
        ``reduction_network.decode``.

        :param torch.Tensor x: Input tensor.
        :return: Solver solution.
        :rtype: torch.Tensor
        """
        reduction_network = self.neural_net["reduction_network"]
        interpolation_network = self.neural_net["interpolation_network"]
        return reduction_network.decode(interpolation_network(x))

    def loss_data(self, input_pts, output_pts):
        """
        The data loss for the ReducedOrderModelSolver solver.
        It computes the loss between
        the network output against the true solution. This function
        should not be override if not intentionally.

        :param LabelTensor input_tensor: The input to the neural networks.
        :param LabelTensor output_tensor: The true solution to compare the
            network solution.
        :return: The residual loss averaged on the input coordinates
        :rtype: torch.Tensor
        """
        # extract networks
        reduction_network = self.neural_net["reduction_network"]
        interpolation_network = self.neural_net["interpolation_network"]
        # encoded representations loss
        encode_repr_inter_net = interpolation_network(input_pts)
        encode_repr_reduction_network = reduction_network.encode(output_pts)
        loss_encode = self.loss(
            encode_repr_inter_net, encode_repr_reduction_network
        )
        # reconstruction loss
        loss_reconstruction = self.loss(
            reduction_network.decode(encode_repr_reduction_network), output_pts
        )

        return loss_encode + loss_reconstruction

    @property
    def neural_net(self):
        """
        Neural network for training. It returns a :obj:`~torch.nn.ModuleDict`
        containing the ``reduction_network`` and ``interpolation_network``.
        """
        return self._neural_net.torchmodel



__all__ = [
    "ContinuousConvBlock",
    "ResidualBlock",
    "EnhancedLinear",
    "SpectralConvBlock1D",
    "SpectralConvBlock2D",
    "SpectralConvBlock3D",
    "FourierBlock1D",
    "FourierBlock2D",
    "FourierBlock3D",
    "PODBlock",
    "PeriodicBoundaryEmbedding",
    "AVNOBlock",
    "LowRankBlock",
]

from .convolution_2d import ContinuousConvBlock
from .residual import ResidualBlock, EnhancedLinear
from .spectral import (
    SpectralConvBlock1D,
    SpectralConvBlock2D,
    SpectralConvBlock3D,
)
from .fourier import FourierBlock1D, FourierBlock2D, FourierBlock3D
from .pod import PODBlock
from .embedding import PeriodicBoundaryEmbedding
from .avno_layer import AVNOBlock
from .lowrank_layer import LowRankBlock



""" Module for PINN """

import sys
from abc import ABCMeta, abstractmethod
import torch

from ...solvers.solver import SolverInterface
from pina.utils import check_consistency
from pina.loss import LossInterface
from pina.problem import InverseProblem
from torch.nn.modules.loss import _Loss

torch.pi = torch.acos(torch.zeros(1)).item() * 2  # which is 3.1415927410125732


class PINNInterface(SolverInterface, metaclass=ABCMeta):
    """
    Base PINN solver class. This class implements the Solver Interface
    for Physics Informed Neural Network solvers.

    This class can be used to
    define PINNs with multiple ``optimizers``, and/or ``models``.
    By default it takes
    an :class:`~pina.problem.abstract_problem.AbstractProblem`, so it is up
    to the user to choose which problem the implemented solver inheriting from
    this class is suitable for.
    """

    def __init__(
        self,
        models,
        problem,
        optimizers,
        optimizers_kwargs,
        extra_features,
        loss,
    ):
        """
        :param models: Multiple torch neural network models instances.
        :type models: list(torch.nn.Module)
        :param problem: A problem definition instance.
        :type problem: AbstractProblem
        :param list(torch.optim.Optimizer) optimizer: A list of neural network
            optimizers to use.
        :param list(dict) optimizer_kwargs: A list of optimizer constructor
            keyword args.
        :param list(torch.nn.Module) extra_features: The additional input
            features to use as augmented input. If ``None`` no extra features
            are passed. If it is a list of :class:`torch.nn.Module`,
            the extra feature list is passed to all models. If it is a list
            of extra features' lists, each single list of extra feature
            is passed to a model.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default :class:`torch.nn.MSELoss`.
        """
        super().__init__(
            models=models,
            problem=problem,
            optimizers=optimizers,
            optimizers_kwargs=optimizers_kwargs,
            extra_features=extra_features,
        )

        # check consistency
        check_consistency(loss, (LossInterface, _Loss), subclass=False)

        # assign variables
        self._loss = loss

        # inverse problem handling
        if isinstance(self.problem, InverseProblem):
            self._params = self.problem.unknown_parameters
            self._clamp_params = self._clamp_inverse_problem_params
        else:
            self._params = None
            self._clamp_params = lambda: None

        # variable used internally to store residual losses at each epoch
        # this variable save the residual at each iteration (not weighted)
        self.__logged_res_losses = []

        # variable used internally in pina for logging. This variable points to
        # the current condition during the training step and returns the
        # condition name. Whenever :meth:`store_log` is called the logged
        # variable will be stored with name = self.__logged_metric
        self.__logged_metric = None

    def training_step(self, batch, _):
        """
        The Physics Informed Solver Training Step. This function takes care
        of the physics informed training step, and it must not be override
        if not intentionally. It handles the batching mechanism, the workload
        division for the various conditions, the inverse problem clamping,
        and loggers.

        :param tuple batch: The batch element in the dataloader.
        :param int batch_idx: The batch index.
        :return: The sum of the loss functions.
        :rtype: LabelTensor
        """

        condition_losses = []
        condition_idx = batch["condition"]

        for condition_id in range(condition_idx.min(), condition_idx.max() + 1):

            condition_name = self._dataloader.condition_names[condition_id]
            condition = self.problem.conditions[condition_name]
            pts = batch["pts"]
            # condition name is logged (if logs enabled)
            self.__logged_metric = condition_name

            if len(batch) == 2:
                samples = pts[condition_idx == condition_id]
                loss = self.loss_phys(samples, condition.equation)
            elif len(batch) == 3:
                samples = pts[condition_idx == condition_id]
                ground_truth = batch["output"][condition_idx == condition_id]
                loss = self.loss_data(samples, ground_truth)
            else:
                raise ValueError("Batch size not supported")

            # add condition losses for each epoch
            condition_losses.append(loss * condition.data_weight)

        # clamp unknown parameters in InverseProblem (if needed)
        self._clamp_params()

        # total loss (must be a torch.Tensor)
        total_loss = sum(condition_losses)
        return total_loss.as_subclass(torch.Tensor)

    def loss_data(self, input_tensor, output_tensor):
        """
        The data loss for the PINN solver. It computes the loss between
        the network output against the true solution. This function
        should not be override if not intentionally.

        :param LabelTensor input_tensor: The input to the neural networks.
        :param LabelTensor output_tensor: The true solution to compare the
            network solution.
        :return: The residual loss averaged on the input coordinates
        :rtype: torch.Tensor
        """
        loss_value = self.loss(self.forward(input_tensor), output_tensor)
        self.store_log(loss_value=float(loss_value))
        return self.loss(self.forward(input_tensor), output_tensor)

    @abstractmethod
    def loss_phys(self, samples, equation):
        """
        Computes the physics loss for the physics informed solver based on given
        samples and equation. This method must be override by all inherited
        classes and it is the core to define a new physics informed solver.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation
            representing the physics.
        :return: The physics loss calculated based on given
            samples and equation.
        :rtype: LabelTensor
        """
        pass

    def compute_residual(self, samples, equation):
        """
        Compute the residual for Physics Informed learning. This function
        returns the :obj:`~pina.equation.equation.Equation` specified in the
        :obj:`~pina.condition.Condition` evaluated at the ``samples`` points.

        :param LabelTensor samples: The samples to evaluate the physics loss.
        :param EquationInterface equation: The governing equation
            representing the physics.
        :return: The residual of the neural network solution.
        :rtype: LabelTensor
        """
        try:
            residual = equation.residual(samples, self.forward(samples))
        except (
            TypeError
        ):  # this occurs when the function has three inputs, i.e. inverse problem
            residual = equation.residual(
                samples, self.forward(samples), self._params
            )
        return residual

    def store_log(self, loss_value):
        """
        Stores the loss value in the logger. This function should be
        called for all conditions. It automatically handles the storing
        conditions names. It must be used
        anytime a specific variable wants to be stored for a specific condition.
        A simple example is to use the variable to store the residual.

        :param str name: The name of the loss.
        :param torch.Tensor loss_value: The value of the loss.
        """
        self.log(
            self.__logged_metric + "_loss",
            loss_value,
            prog_bar=True,
            logger=True,
            on_epoch=True,
            on_step=False,
        )
        self.__logged_res_losses.append(loss_value)

    def on_train_epoch_end(self):
        """
        At the end of each epoch we free the stored losses. This function
        should not be override if not intentionally.
        """
        if self.__logged_res_losses:
            # storing mean loss
            self.__logged_metric = "mean"
            self.store_log(
                sum(self.__logged_res_losses) / len(self.__logged_res_losses)
            )
            # free the logged losses
            self.__logged_res_losses = []
        return super().on_train_epoch_end()

    def _clamp_inverse_problem_params(self):
        """
        Clamps the parameters of the inverse problem
        solver to the specified ranges.
        """
        for v in self._params:
            self._params[v].data.clamp_(
                self.problem.unknown_parameter_domain.range_[v][0],
                self.problem.unknown_parameter_domain.range_[v][1],
            )

    @property
    def loss(self):
        """
        Loss used for training.
        """
        return self._loss

    @property
    def current_condition_name(self):
        """
        Returns the condition name. This function can be used inside the
        :meth:`loss_phys` to extract the condition at which the loss is
        computed.
        """
        return self.__logged_metric



__all__ = [
    "PINNInterface",
    "PINN",
    "GPINN",
    "CausalPINN",
    "CompetitivePINN",
    "SAPINN",
]

from .basepinn import PINNInterface
from .pinn import PINN
from .gpinn import GPINN
from .causalpinn import CausalPINN
from .competitive_pinn import CompetitivePINN
from .sapinn import SAPINN



__all__ = ["SwitchOptimizer", "R3Refinement", "MetricTracker"]

from .optimizer_callbacks import SwitchOptimizer
from .adaptive_refinment_callbacks import R3Refinement
from .processing_callbacks import MetricTracker



__all__ = [
    "PINN",
    "Trainer",
    "LabelTensor",
    "Plotter",
    "Condition",
    "SamplePointDataset",
    "SamplePointLoader",
]

from .meta import *
from .label_tensor import LabelTensor
from .solvers.solver import SolverInterface
from .trainer import Trainer
from .plotter import Plotter
from .condition import Condition
from .dataset import SamplePointDataset
from .dataset import SamplePointLoader



__all__ = [
    "__project__",
    "__title__",
    "__author__",
    "__copyright__",
    "__license__",
    "__version__",
    "__mail__",
    "__maintainer__",
    "__status__",
]

__project__ = "PINA"
__title__ = "pina"
__author__ = "PINA Contributors"
__copyright__ = "2021-2024, PINA Contributors"
__license__ = "MIT"
__version__ = "0.1.1"
__mail__ = "demo.nicola@gmail.com, dario.coscia@sissa.it"  # TODO
__maintainer__ = __author__
__status__ = "Alpha"
__packagename__ = "pina-mathlab"



__all__ = [
    "SolverInterface",
    "PINNInterface",
    "PINN",
    "GPINN",
    "CausalPINN",
    "CompetitivePINN",
    "SAPINN",
    "SupervisedSolver",
    "ReducedOrderModelSolver",
    "GAROM",
]

from .solver import SolverInterface
from .pinns import *
from .supervised import SupervisedSolver
from .rom import ReducedOrderModelSolver
from .garom import GAROM



""" Solver module. """

from abc import ABCMeta, abstractmethod
from ..model.network import Network
import pytorch_lightning
from ..utils import check_consistency
from ..problem import AbstractProblem
import torch
import sys


class SolverInterface(pytorch_lightning.LightningModule, metaclass=ABCMeta):
    """
    Solver base class. This class inherits is a wrapper of
    LightningModule class, inheriting all the
    LightningModule methods.
    """

    def __init__(
        self,
        models,
        problem,
        optimizers,
        optimizers_kwargs,
        extra_features=None,
    ):
        """
        :param models: A torch neural network model instance.
        :type models: torch.nn.Module
        :param problem: A problem definition instance.
        :type problem: AbstractProblem
        :param list(torch.optim.Optimizer) optimizer: A list of neural network optimizers to
            use.
        :param list(dict) optimizer_kwargs: A list of optimizer constructor keyword args.
        :param list(torch.nn.Module) extra_features: The additional input
            features to use as augmented input. If ``None`` no extra features
            are passed. If it is a list of :class:`torch.nn.Module`, the extra feature
            list is passed to all models. If it is a list of extra features' lists,
            each single list of extra feature is passed to a model.
        """
        super().__init__()

        # check consistency of the inputs
        check_consistency(models, torch.nn.Module)
        check_consistency(problem, AbstractProblem)
        check_consistency(optimizers, torch.optim.Optimizer, subclass=True)
        check_consistency(optimizers_kwargs, dict)

        # put everything in a list if only one input
        if not isinstance(models, list):
            models = [models]
        if not isinstance(optimizers, list):
            optimizers = [optimizers]
            optimizers_kwargs = [optimizers_kwargs]

        # number of models and optimizers
        len_model = len(models)
        len_optimizer = len(optimizers)
        len_optimizer_kwargs = len(optimizers_kwargs)

        # check length consistency optimizers
        if len_model != len_optimizer:
            raise ValueError(
                "You must define one optimizer for each model."
                f"Got {len_model} models, and {len_optimizer}"
                " optimizers."
            )

        # check length consistency optimizers kwargs
        if len_optimizer_kwargs != len_optimizer:
            raise ValueError(
                "You must define one dictionary of keyword"
                " arguments for each optimizers."
                f"Got {len_optimizer} optimizers, and"
                f" {len_optimizer_kwargs} dicitionaries"
            )

        # extra features handling
        if (extra_features is None) or (len(extra_features) == 0):
            extra_features = [None] * len_model
        else:
            # if we only have a list of extra features
            if not isinstance(extra_features[0], (tuple, list)):
                extra_features = [extra_features] * len_model
            else:  # if we have a list of list extra features
                if len(extra_features) != len_model:
                    raise ValueError(
                        "You passed a list of extrafeatures list with len"
                        f"different of models len. Expected {len_model} "
                        f"got {len(extra_features)}. If you want to use "
                        "the same list of extra features for all models, "
                        "just pass a list of extrafeatures and not a list "
                        "of list of extra features."
                    )

        # assigning model and optimizers
        self._pina_models = []
        self._pina_optimizers = []

        for idx in range(len_model):
            model_ = Network(
                model=models[idx],
                input_variables=problem.input_variables,
                output_variables=problem.output_variables,
                extra_features=extra_features[idx],
            )
            optim_ = optimizers[idx](
                model_.parameters(), **optimizers_kwargs[idx]
            )
            self._pina_models.append(model_)
            self._pina_optimizers.append(optim_)

        # assigning problem
        self._pina_problem = problem

    @abstractmethod
    def forward(self, *args, **kwargs):
        pass

    @abstractmethod
    def training_step(self):
        pass

    @abstractmethod
    def configure_optimizers(self):
        pass

    @property
    def models(self):
        """
        The torch model."""
        return self._pina_models

    @property
    def optimizers(self):
        """
        The torch model."""
        return self._pina_optimizers

    @property
    def problem(self):
        """
        The problem formulation."""
        return self._pina_problem

    def on_train_start(self):
        """
        On training epoch start this function is call to do global checks for
        the different solvers.
        """

        # 1. Check the verison for dataloader
        dataloader = self.trainer.train_dataloader
        if sys.version_info < (3, 8):
            dataloader = dataloader.loaders
        self._dataloader = dataloader

        return super().on_train_start()

    # @model.setter
    # def model(self, new_model):
    #     """
    #     Set the torch."""
    #     check_consistency(new_model, nn.Module, 'torch model')
    #     self._model= new_model

    # @problem.setter
    # def problem(self, problem):
    #     """
    #     Set the problem formulation."""
    #     check_consistency(problem, AbstractProblem, 'pina problem')
    #     self._problem = problem
